"""
This module encapsulates available sensor types.
"""

from typing import Optional, List

import redvox.api1000.common.common as common
import redvox.api1000.common.typing
import redvox.api1000.proto.redvox_api_m_pb2 as redvox_api_m_pb2
import redvox.api1000.common.generic
import redvox.api1000.wrapped_redvox_packet.sensors.audio as audio
import redvox.api1000.wrapped_redvox_packet.sensors.image as image
import redvox.api1000.wrapped_redvox_packet.sensors.location as location
import redvox.api1000.wrapped_redvox_packet.sensors.single as single
import redvox.api1000.wrapped_redvox_packet.sensors.xyz as xyz


class Sensors(
    redvox.api1000.common.generic.ProtoBase[redvox_api_m_pb2.RedvoxPacketM.Sensors]
):
    """
    This class encapsulated available API M sensors.
    """

    # These are used for checking if a field is present or not
    __ACCELEROMETER_FIELD_NAME: str = "accelerometer"
    __AMBIENT_TEMPERATURE_FIELD_NAME: str = "ambient_temperature"
    __AUDIO_FIELD_NAME: str = "audio"
    __COMPRESSED_AUDIO_FIELD_NAME: str = "compressed_audio"
    __GRAVITY_FIELD_NAME: str = "gravity"
    __GYROSCOPE_FIELD_NAME: str = "gyroscope"
    __IMAGE_FIELD_NAME: str = "image"
    __LIGHT_FIELD_NAME: str = "light"
    __LINEAR_ACCELERATION_FIELD_NAME: str = "linear_acceleration"
    __LOCATION_FIELD_NAME: str = "location"
    __MAGNETOMETER_FIELD_NAME: str = "magnetometer"
    __ORIENTATION_FIELD_NAME: str = "orientation"
    __PRESSURE_FIELD_NAME: str = "pressure"
    __PROXIMITY_FIELD_NAME: str = "proximity"
    __RELATIVE_HUMIDITY_FIELD_NAME: str = "relative_humidity"
    __ROTATION_VECTOR: str = "rotation_vector"
    __VELOCITY: str = "velocity"

    def __init__(self, sensors_proto: redvox_api_m_pb2.RedvoxPacketM.Sensors):
        super().__init__(sensors_proto)
        self._accelerometer: xyz.Xyz = xyz.Xyz(sensors_proto.accelerometer)
        self._ambient_temperature: single.Single = single.Single(
            sensors_proto.ambient_temperature
        )
        self._audio: audio.Audio = audio.Audio(sensors_proto.audio)
        self._compressed_audio: audio.CompressedAudio = audio.CompressedAudio(
            sensors_proto.compressed_audio
        )
        self._gravity: xyz.Xyz = xyz.Xyz(sensors_proto.gravity)
        self._gyroscope: xyz.Xyz = xyz.Xyz(sensors_proto.gyroscope)
        self._image: image.Image = image.Image(sensors_proto.image)
        self._light: single.Single = single.Single(sensors_proto.light)
        self._linear_acceleration: xyz.Xyz = xyz.Xyz(sensors_proto.linear_acceleration)
        self._location: location.Location = location.Location(sensors_proto.location)
        self._magnetometer: xyz.Xyz = xyz.Xyz(sensors_proto.magnetometer)
        self._orientation: xyz.Xyz = xyz.Xyz(sensors_proto.orientation)
        self._pressure: single.Single = single.Single(sensors_proto.pressure)
        self._proximity: single.Single = single.Single(sensors_proto.proximity)
        self._relative_humidity: single.Single = single.Single(
            sensors_proto.relative_humidity
        )
        self._rotation_vector: xyz.Xyz = xyz.Xyz(sensors_proto.rotation_vector)
        self._velocity: xyz.Xyz = xyz.Xyz(sensors_proto.velocity)

    @staticmethod
    def new() -> "Sensors":
        """
        :return: A new, empty Sensors instance
        """
        return Sensors(redvox_api_m_pb2.RedvoxPacketM.Sensors())

    def has_accelerometer(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__ACCELEROMETER_FIELD_NAME)

    def get_accelerometer(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._accelerometer if self.has_accelerometer() else None

    def new_accelerometer(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_accelerometer()
        self.get_proto().accelerometer.SetInParent()
        self._accelerometer = xyz.Xyz(self.get_proto().accelerometer)
        self._accelerometer.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._accelerometer.set_unit_xyz(common.Unit.METERS_PER_SECOND_SQUARED)
        return self._accelerometer

    def set_accelerometer(self, accelerometer: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param accelerometer: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(accelerometer, [xyz.Xyz])
        self.get_proto().accelerometer.CopyFrom(accelerometer.get_proto())
        self._accelerometer = xyz.Xyz(self.get_proto().accelerometer)
        return self

    def remove_accelerometer(self) -> "Sensors":
        """
        Removes this channel
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__ACCELEROMETER_FIELD_NAME)
        return self

    def validate_accelerometer(self) -> bool:
        """
        Checks if there are no errors with the accelerometer and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(
                xyz.validate_xyz(
                    self._accelerometer, common.Unit.METERS_PER_SECOND_SQUARED
                )
            )
            < 1
            and self._accelerometer.get_x_samples().get_values_count() > 0
        )

    def has_ambient_temperature(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__AMBIENT_TEMPERATURE_FIELD_NAME)

    def get_ambient_temperature(self) -> Optional[single.Single]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._ambient_temperature if self.has_ambient_temperature() else None

    def new_ambient_temperature(self) -> single.Single:
        """
        :return: A new empty sensor
        """
        self.remove_ambient_temperature()
        self.get_proto().ambient_temperature.SetInParent()
        self._ambient_temperature = single.Single(self.get_proto().ambient_temperature)
        self._ambient_temperature.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._ambient_temperature.get_samples().set_unit(common.Unit.DEGREES_CELSIUS)
        return self._ambient_temperature

    def set_ambient_temperature(self, ambient_temperature: single.Single) -> "Sensors":
        """
        Sets the channel
        :param ambient_temperature: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(ambient_temperature, [single.Single])
        self.get_proto().ambient_temperature.CopyFrom(ambient_temperature.get_proto())
        self._ambient_temperature = single.Single(self.get_proto().ambient_temperature)
        return self

    def remove_ambient_temperature(self) -> "Sensors":
        """
        Removes this channel
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__AMBIENT_TEMPERATURE_FIELD_NAME)
        return self

    def validate_ambient_temperature(self) -> bool:
        """
        Checks if there are no errors with the ambient temperature sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(
                single.validate_single(
                    self._ambient_temperature, common.Unit.DEGREES_CELSIUS
                )
            )
            < 1
            and self._ambient_temperature.get_samples().get_values_count() > 0
        )

    def has_audio(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__AUDIO_FIELD_NAME)

    def get_audio(self) -> Optional[audio.Audio]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._audio if self.has_audio() else None

    def new_audio(self) -> audio.Audio:
        """
        :return: A new empty sensor
        """
        self.remove_audio()
        self.get_proto().audio.SetInParent()
        self._audio = audio.Audio(self.get_proto().audio)
        # noinspection PyTypeChecker
        self._audio.get_samples().set_unit(common.Unit.LSB_PLUS_MINUS_COUNTS)
        return self._audio

    def set_audio(self, _audio: audio.Audio) -> "Sensors":
        """
        Sets the channel
        :param _audio: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(_audio, [audio.Audio])
        self.get_proto().audio.CopyFrom(_audio.get_proto())
        self._audio = audio.Audio(self.get_proto().audio)
        return self

    def remove_audio(self) -> "Sensors":
        """
        Removes this channel
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__AUDIO_FIELD_NAME)
        return self

    def validate_audio(self) -> bool:
        """
        Checks if there are no errors with the audio sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        return (
            len(audio.validate_audio(self._audio)) < 1
            and self._audio.get_samples().get_values_count() > 0
        )

    def has_compressed_audio(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__COMPRESSED_AUDIO_FIELD_NAME)

    def get_compressed_audio(self) -> Optional[audio.CompressedAudio]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._compressed_audio if self.has_compressed_audio() else None

    def new_compressed_audio(self) -> audio.CompressedAudio:
        """
        :return: A new empty sensor
        """
        self.remove_compressed_audio()
        self.get_proto().compressed_audio.SetInParent()
        self._compressed_audio = audio.CompressedAudio(
            self.get_proto().compressed_audio
        )
        return self._compressed_audio

    def set_compressed_audio(
        self, compressed_audio: audio.CompressedAudio
    ) -> "Sensors":
        """
        Sets the channel
        :param compressed_audio: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(
            compressed_audio, [audio.CompressedAudio]
        )
        # noinspection Mypy
        self.get_proto().compressed_audio.CopyFrom(compressed_audio.get_proto())
        self._compressed_audio = audio.CompressedAudio(
            self.get_proto().compressed_audio
        )
        return self

    def remove_compressed_audio(self) -> "Sensors":
        """
        Removes this channel
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__COMPRESSED_AUDIO_FIELD_NAME)
        return self

    def validate_compressed_audio(self) -> bool:
        """
        Checks if there are no errors with the compressed audio sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        return len(audio.validate_compressed_audio(self._compressed_audio)) < 1

    def has_gravity(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker,Mypy
        return self.get_proto().HasField(Sensors.__GRAVITY_FIELD_NAME)

    def get_gravity(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._gravity if self.has_gravity() else None

    def new_gravity(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_gravity()
        self.get_proto().gravity.SetInParent()
        self._gravity = xyz.Xyz(self.get_proto().gravity)
        self._gravity.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._gravity.set_unit_xyz(common.Unit.METERS_PER_SECOND_SQUARED)
        return self._gravity

    def set_gravity(self, gravity: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param gravity: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(gravity, [xyz.Xyz])
        self.get_proto().gravity.CopyFrom(gravity.get_proto())
        self._gravity = xyz.Xyz(self.get_proto().gravity)
        return self

    def remove_gravity(self) -> "Sensors":
        """
        Removes this channel
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__GRAVITY_FIELD_NAME)
        return self

    def validate_gravity(self) -> bool:
        """
        Checks if there are no errors with the gravity sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(xyz.validate_xyz(self._gravity, common.Unit.METERS_PER_SECOND_SQUARED))
            < 1
            and self._gravity.get_x_samples().get_values_count() > 0
        )

    def has_gyroscope(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__GYROSCOPE_FIELD_NAME)

    def get_gyroscope(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._gyroscope if self.has_gyroscope() else None

    def new_gyroscope(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_gyroscope()
        self.get_proto().gyroscope.SetInParent()
        self._gyroscope = xyz.Xyz(self.get_proto().gyroscope)
        self._gyroscope.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._gyroscope.set_unit_xyz(common.Unit.RADIANS_PER_SECOND)
        return self._gyroscope

    def set_gyroscope(self, gyroscope: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param gyroscope: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(gyroscope, [xyz.Xyz])
        self.get_proto().gyroscope.CopyFrom(gyroscope.get_proto())
        self._gyroscope = xyz.Xyz(self.get_proto().gyroscope)
        return self

    def remove_gyroscope(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__GYROSCOPE_FIELD_NAME)
        return self

    def validate_gyroscope(self) -> bool:
        """
        Checks if there are no errors with the gyroscope and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(xyz.validate_xyz(self._gyroscope, common.Unit.RADIANS_PER_SECOND)) < 1
            and self._gyroscope.get_x_samples().get_values_count() > 0
        )

    def has_image(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__IMAGE_FIELD_NAME)

    def get_image(self) -> Optional[image.Image]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._image if self.has_image() else None

    def new_image(self) -> image.Image:
        """
        :return: A new empty sensor
        """
        self.remove_image()
        self.get_proto().image.SetInParent()
        self._image = image.Image(self.get_proto().image)
        # noinspection PyTypeChecker
        self._image.set_image_codec(image.ImageCodec.PNG)
        return self._image

    def set_image(self, _image: image.Image) -> "Sensors":
        """
        Sets the channel
        :param _image: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(_image, [image.Image])
        self.get_proto().image.CopyFrom(_image.get_proto())
        self._image = image.Image(self.get_proto().image)
        return self

    def remove_image(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__IMAGE_FIELD_NAME)
        return self

    def validate_image(self) -> bool:
        """
        Checks if there are no errors with the image sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        return (
            len(image.validate_image(self._image)) < 1
            and self._image.get_num_images() > 0
        )

    def has_light(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__LIGHT_FIELD_NAME)

    def get_light(self) -> Optional[single.Single]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._light if self.has_light() else None

    def new_light(self) -> single.Single:
        """
        :return: A new empty sensor
        """
        self.remove_light()
        self.get_proto().light.SetInParent()
        self._light = single.Single(self.get_proto().light)
        self._light.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._light.get_samples().set_unit(common.Unit.LUX)
        return self._light

    def set_light(self, light: single.Single) -> "Sensors":
        """
        Sets the channel
        :param light: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(light, [single.Single])
        self.get_proto().light.CopyFrom(light.get_proto())
        self._light = single.Single(self.get_proto().light)
        return self

    def remove_light(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__LIGHT_FIELD_NAME)
        return self

    def validate_light(self) -> bool:
        """
        Checks if there are no errors with the light sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(single.validate_single(self._light, common.Unit.LUX)) < 1
            and self._light.get_samples().get_values_count() > 0
        )

    def has_linear_acceleration(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__LINEAR_ACCELERATION_FIELD_NAME)

    def get_linear_acceleration(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._linear_acceleration if self.has_linear_acceleration() else None

    def new_linear_acceleration(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_linear_acceleration()
        self.get_proto().linear_acceleration.SetInParent()
        self._linear_acceleration = xyz.Xyz(self.get_proto().linear_acceleration)
        self._linear_acceleration.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._linear_acceleration.set_unit_xyz(common.Unit.METERS_PER_SECOND_SQUARED)
        return self._linear_acceleration

    def set_linear_acceleration(self, linear_acceleration: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param linear_acceleration: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(linear_acceleration, [xyz.Xyz])
        self.get_proto().linear_acceleration.CopyFrom(linear_acceleration.get_proto())
        self._linear_acceleration = xyz.Xyz(self.get_proto().linear_acceleration)
        return self

    def remove_linear_acceleration(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__LINEAR_ACCELERATION_FIELD_NAME)
        return self

    def validate_linear_acceleration(self) -> bool:
        """
        Checks if there are no errors with the linear acceleration sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(
                xyz.validate_xyz(
                    self._linear_acceleration, common.Unit.METERS_PER_SECOND_SQUARED
                )
            )
            < 1
            and self._linear_acceleration.get_x_samples().get_values_count() > 0
        )

    def has_location(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__LOCATION_FIELD_NAME)

    def get_location(self) -> Optional[location.Location]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._location if self.has_location() else None

    def new_location(self) -> location.Location:
        """
        :return: A new empty sensor
        """
        self.remove_location()
        self.get_proto().location.SetInParent()
        self._location = location.Location(self.get_proto().location)
        self._location.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._location.get_latitude_samples().set_unit(common.Unit.DECIMAL_DEGREES)
        # noinspection PyTypeChecker
        self._location.get_longitude_samples().set_unit(common.Unit.DECIMAL_DEGREES)
        # noinspection PyTypeChecker
        self._location.get_altitude_samples().set_unit(common.Unit.METERS)
        # noinspection PyTypeChecker
        self._location.get_speed_samples().set_unit(common.Unit.METERS_PER_SECOND)
        # noinspection PyTypeChecker
        self._location.get_bearing_samples().set_unit(common.Unit.DECIMAL_DEGREES)
        # noinspection PyTypeChecker
        self._location.get_horizontal_accuracy_samples().set_unit(common.Unit.METERS)
        # noinspection PyTypeChecker
        self._location.get_vertical_accuracy_samples().set_unit(common.Unit.METERS)
        # noinspection PyTypeChecker
        self._location.get_speed_accuracy_samples().set_unit(
            common.Unit.METERS_PER_SECOND
        )
        # noinspection PyTypeChecker
        self._location.get_bearing_accuracy_samples().set_unit(
            common.Unit.DECIMAL_DEGREES
        )
        return self._location

    def set_location(self, _location: location.Location) -> "Sensors":
        """
        Sets the channel
        :param _location: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(_location, [location.Location])
        self.get_proto().location.CopyFrom(_location.get_proto())
        self._location = location.Location(self.get_proto().location)
        return self

    def remove_location(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__LOCATION_FIELD_NAME)
        return self

    def validate_location(self) -> bool:
        """
        Checks if there are no errors with the location
        :return: True if no errors
        """
        return len(location.validate_location(self._location)) < 1

    def has_magnetometer(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__MAGNETOMETER_FIELD_NAME)

    def get_magnetometer(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._magnetometer if self.has_magnetometer() else None

    def new_magnetometer(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_magnetometer()
        self.get_proto().magnetometer.SetInParent()
        self._magnetometer = xyz.Xyz(self.get_proto().magnetometer)
        self._magnetometer.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._magnetometer.set_unit_xyz(common.Unit.MICROTESLA)
        return self._magnetometer

    def set_magnetometer(self, magnetometer: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param magnetometer: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(magnetometer, [xyz.Xyz])
        self.get_proto().magnetometer.CopyFrom(magnetometer.get_proto())
        self._magnetometer = xyz.Xyz(self.get_proto().magnetometer)
        return self

    def remove_magnetometer(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__MAGNETOMETER_FIELD_NAME)
        return self

    def validate_magnetometer(self) -> bool:
        """
        Checks if there are no errors with the magnetometer and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(xyz.validate_xyz(self._magnetometer, common.Unit.MICROTESLA)) < 1
            and self._magnetometer.get_x_samples().get_values_count() > 0
        )

    def has_orientation(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__ORIENTATION_FIELD_NAME)

    def get_orientation(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._orientation if self.has_orientation() else None

    def new_orientation(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_orientation()
        self.get_proto().orientation.SetInParent()
        self._orientation = xyz.Xyz(self.get_proto().orientation)
        self._orientation.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._orientation.set_unit_xyz(common.Unit.RADIANS)
        return self._orientation

    def set_orientation(self, orientation: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param orientation: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(orientation, [xyz.Xyz])
        self.get_proto().orientation.CopyFrom(orientation.get_proto())
        self._orientation = xyz.Xyz(self.get_proto().orientation)
        return self

    def remove_orientation(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__ORIENTATION_FIELD_NAME)
        return self

    def validate_orientation(self) -> bool:
        """
        Checks if there are no errors with the orientation sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(xyz.validate_xyz(self._orientation, common.Unit.RADIANS)) < 1
            and self._orientation.get_x_samples().get_values_count() > 0
        )

    def has_pressure(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__PRESSURE_FIELD_NAME)

    def get_pressure(self) -> Optional[single.Single]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._pressure if self.has_pressure() else None

    def new_pressure(self) -> single.Single:
        """
        :return: A new empty sensor
        """
        self.remove_pressure()
        self.get_proto().pressure.SetInParent()
        self._pressure = single.Single(self.get_proto().pressure)
        self._pressure.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._pressure.get_samples().set_unit(common.Unit.KILOPASCAL)
        return self._pressure

    def set_pressure(self, pressure: single.Single) -> "Sensors":
        """
        Sets the channel
        :param pressure: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(pressure, [single.Single])
        self.get_proto().pressure.CopyFrom(pressure.get_proto())
        self._pressure = single.Single(self.get_proto().pressure)
        return self

    def remove_pressure(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__PRESSURE_FIELD_NAME)
        return self

    def validate_pressure(self) -> bool:
        """
        Checks if there are no errors with the pressure sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(single.validate_single(self._pressure, common.Unit.KILOPASCAL)) < 1
            and self._pressure.get_samples().get_values_count() > 0
        )

    def has_proximity(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__PROXIMITY_FIELD_NAME)

    def get_proximity(self) -> Optional[single.Single]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._proximity if self.has_proximity() else None

    def new_proximity(self) -> single.Single:
        """
        :return: A new empty sensor
        """
        self.remove_proximity()
        self.get_proto().proximity.SetInParent()
        self._proximity = single.Single(self.get_proto().proximity)
        self._proximity.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._proximity.get_samples().set_unit(common.Unit.CENTIMETERS)
        return self._proximity

    def set_proximity(self, proximity: single.Single) -> "Sensors":
        """
        Sets the channel
        :param proximity: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(proximity, [single.Single])
        self.get_proto().proximity.CopyFrom(proximity.get_proto())
        self._proximity = single.Single(self.get_proto().proximity)
        return self

    def remove_proximity(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__PROXIMITY_FIELD_NAME)
        return self

    def validate_proximity(self) -> bool:
        """
        Checks if there are no errors with the proximity sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(single.validate_single(self._proximity, common.Unit.CENTIMETERS)) < 1
            and self._proximity.get_samples().get_values_count() > 0
        )

    def has_relative_humidity(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__RELATIVE_HUMIDITY_FIELD_NAME)

    def get_relative_humidity(self) -> Optional[single.Single]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._relative_humidity if self.has_relative_humidity() else None

    def new_relative_humidity(self) -> single.Single:
        """
        :return: A new empty sensor
        """
        self.remove_relative_humidity()
        self.get_proto().relative_humidity.SetInParent()
        self._relative_humidity = single.Single(self.get_proto().relative_humidity)
        self._relative_humidity.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._relative_humidity.get_samples().set_unit(common.Unit.PERCENTAGE)
        return self._relative_humidity

    def set_relative_humidity(self, relative_humidity: single.Single) -> "Sensors":
        """
        Sets the channel
        :param relative_humidity: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(relative_humidity, [single.Single])
        self.get_proto().relative_humidity.CopyFrom(relative_humidity.get_proto())
        self._relative_humidity = single.Single(self.get_proto().relative_humidity)
        return self

    def remove_relative_humidity(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__RELATIVE_HUMIDITY_FIELD_NAME)
        return self

    def validate_relative_humidity(self) -> bool:
        """
        Checks if there are no errors with the relative_humidity sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(single.validate_single(self._relative_humidity, common.Unit.PERCENTAGE))
            < 1
            and self._relative_humidity.get_samples().get_values_count() > 0
        )

    def has_rotation_vector(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__ROTATION_VECTOR)

    def get_rotation_vector(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._rotation_vector if self.has_rotation_vector() else None

    def new_rotation_vector(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_rotation_vector()
        self.get_proto().rotation_vector.SetInParent()
        self._rotation_vector = xyz.Xyz(self.get_proto().rotation_vector)
        self._rotation_vector.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._rotation_vector.set_unit_xyz(common.Unit.UNITLESS)
        return self._rotation_vector

    def set_rotation_vector(self, rotation_vector: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param rotation_vector: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(rotation_vector, [xyz.Xyz])
        self.get_proto().rotation_vector.CopyFrom(rotation_vector.get_proto())
        self._rotation_vector = xyz.Xyz(self.get_proto().rotation_vector)
        return self

    def remove_rotation_vector(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__ROTATION_VECTOR)
        return self

    def validate_rotation_vector(self) -> bool:
        """
        Checks if there are no errors with the rotation vector sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(xyz.validate_xyz(self._rotation_vector, common.Unit.UNITLESS)) < 1
            and self._rotation_vector.get_x_samples().get_values_count() > 0
        )

    def has_velocity(self) -> bool:
        """
        :return: If this packet contains this channel
        """
        # noinspection PyTypeChecker
        return self.get_proto().HasField(Sensors.__VELOCITY)

    def get_velocity(self) -> Optional[xyz.Xyz]:
        """
        :return: The this sensor if it exists otherwise None
        """
        return self._velocity if self.has_velocity() else None

    def new_velocity(self) -> xyz.Xyz:
        """
        :return: A new empty sensor
        """
        self.remove_velocity()
        self.get_proto().velocity.SetInParent()
        self._velocity = xyz.Xyz(self.get_proto().velocity)
        self._velocity.get_timestamps().set_default_unit()
        # noinspection PyTypeChecker
        self._velocity.set_unit_xyz(common.Unit.METERS_PER_SECOND)
        return self._velocity

    def set_velocity(self, velocity: xyz.Xyz) -> "Sensors":
        """
        Sets the channel
        :param velocity: Channel to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(velocity, [xyz.Xyz])
        self.get_proto().velocity.CopyFrom(velocity.get_proto())
        self._velocity = xyz.Xyz(self.get_proto().velocity)
        return self

    def remove_velocity(self) -> "Sensors":
        """
        Removes this sensor
        :return: A modified instance of self
        """
        # noinspection PyTypeChecker
        self.get_proto().ClearField(Sensors.__VELOCITY)
        return self

    def validate_velocity(self) -> bool:
        """
        Checks if there are no errors with the sensor and it contains at least 1 data entry
        :return: True if no errors
        """
        # noinspection PyTypeChecker
        return (
            len(xyz.validate_xyz(self._velocity, common.Unit.METERS_PER_SECOND)) < 1
            and self._velocity.get_x_samples().get_values_count() > 0
        )


def validate_sensors(sensors_list: Sensors) -> List[str]:
    """
    Validates sensors.
    :param sensors_list: Sensors to validate
    :return: A list of validated errors
    """
    # audio is the only sensor that every packet must have
    errors_list = []
    if not sensors_list.has_audio() and not sensors_list.has_compressed_audio():
        errors_list.append("Sensors list missing audio sensor")
    else:
        if sensors_list.has_audio():
            errors_list.extend(audio.validate_audio(sensors_list.get_audio()))
        if sensors_list.has_compressed_audio():
            errors_list.extend(
                audio.validate_compressed_audio(sensors_list.get_compressed_audio())
            )
    if sensors_list.has_accelerometer():
        errors_list.extend(
            xyz.validate_xyz(
                sensors_list.get_accelerometer(), common.Unit.METERS_PER_SECOND_SQUARED
            )
        )
    if sensors_list.has_ambient_temperature():
        errors_list.extend(
            single.validate_single(
                sensors_list.get_ambient_temperature(), common.Unit.DEGREES_CELSIUS
            )
        )
    if sensors_list.has_gravity():
        errors_list.extend(
            xyz.validate_xyz(
                sensors_list.get_gravity(), common.Unit.METERS_PER_SECOND_SQUARED
            )
        )
    if sensors_list.has_gyroscope():
        errors_list.extend(
            xyz.validate_xyz(
                sensors_list.get_gyroscope(), common.Unit.RADIANS_PER_SECOND
            )
        )
    if sensors_list.has_image():
        errors_list.extend(image.validate_image(sensors_list.get_image()))
    if sensors_list.has_light():
        errors_list.extend(
            single.validate_single(sensors_list.get_light(), common.Unit.LUX)
        )
    if sensors_list.has_linear_acceleration():
        errors_list.extend(
            xyz.validate_xyz(
                sensors_list.get_linear_acceleration(),
                common.Unit.METERS_PER_SECOND_SQUARED,
            )
        )
    if sensors_list.has_location():
        errors_list.extend(location.validate_location(sensors_list.get_location()))
    if sensors_list.has_magnetometer():
        errors_list.extend(
            xyz.validate_xyz(sensors_list.get_magnetometer(), common.Unit.MICROTESLA)
        )
    if sensors_list.has_orientation():
        errors_list.extend(
            xyz.validate_xyz(sensors_list.get_orientation(), common.Unit.RADIANS)
        )
    if sensors_list.has_pressure():
        errors_list.extend(
            single.validate_single(sensors_list.get_pressure(), common.Unit.KILOPASCAL)
        )
    if sensors_list.has_proximity():
        errors_list.extend(
            single.validate_single(
                sensors_list.get_proximity(), common.Unit.CENTIMETERS
            )
        )
    if sensors_list.has_relative_humidity():
        errors_list.extend(
            single.validate_single(
                sensors_list.get_relative_humidity(), common.Unit.PERCENTAGE
            )
        )
    if sensors_list.has_rotation_vector():
        errors_list.extend(
            xyz.validate_xyz(sensors_list.get_rotation_vector(), common.Unit.UNITLESS)
        )
    if sensors_list.has_velocity():
        errors_list.extend(
            xyz.validate_xyz(sensors_list.get_velocity(), common.Unit.METERS_PER_SECOND)
        )
    return errors_list



"""
This module contains classes and function that make use of generics. As of April 2020, IntelliJ still does not parse
generics correctly and cause errors to appear where there are none. However, generics can still be statically checked by
mypy, and unless we decide to remove generics, we will place them here.
"""

from typing import Generic, Callable, List, Dict, TypeVar

from google.protobuf.json_format import MessageToJson, MessageToDict

from redvox.api1000.common.metadata import Metadata
from redvox.api1000.common.lz4 import compress

# pylint: disable=C0103
T = TypeVar("T")  # Type parameter for transformed wrapper type
# pylint: disable=C0103
P = TypeVar("P")  # Type parameter for protobuf type


class ProtoRepeatedMessage(Generic[P, T]):
    """
    Encapsulates protobuf repeated fields while transforming values between protobuf and wrapper types.
    """

    def __init__(
        self,
        parent_proto,
        repeated_field_proto,
        repeated_field_name: str,
        from_proto: Callable[[P], T],
        to_proto: Callable[[T], P],
    ):
        """
        :param parent_proto: A reference to this message's parent protobuf
        :param repeated_field_proto: A reference to the repeated field protobuf
        :param repeated_field_name: The name of the repeated protobuf field
        :param from_proto: A function that converts a repeated protobuf value into its wrapper equivelent
        :param to_proto: A function that converts a repeated wrapper value into its protobuf equivelent
        """
        self._parent_proto = parent_proto
        self._repeated_field_proto = repeated_field_proto
        self._repeated_field_name = repeated_field_name
        self._from_proto: Callable[[P], T] = from_proto
        self._to_proto: Callable[[T], P] = to_proto

    def get_count(self) -> int:
        """
        Returns the number of items in this repeated collection.
        :return: The number of items in this repeated collection.
        """
        return len(self._repeated_field_proto)

    def get_values(self) -> List[T]:
        """
        Returns a list of wrapped items stored in this collection.
        :return: A list of wrapped items stored in this collection.
        """
        return list(map(self._from_proto, self._repeated_field_proto))

    def set_values(self, values: List[T]) -> "ProtoRepeatedMessage[P, T]":
        """
        Sets the contents of this collection to the passed in values. The wrapped values are automatically converted
        into protobuf messages.
        :param values: The wrapped values to set.
        :return: This instance of ProtoRepeatedMessage.
        """
        return self.clear_values().append_values(values)

    def append_values(self, values: List[T]) -> "ProtoRepeatedMessage[P, T]":
        """
        Appends the given values to the collection, automatically converting them into protobuf messages.
        :param values: The wrapped values to append.
        :return: This instance of ProtoRepeatedMessage.
        """
        self._repeated_field_proto.extend(list(map(self._to_proto, values)))
        return self

    def clear_values(self) -> "ProtoRepeatedMessage[P, T]":
        """
        Clears all values in this collection.
        :return: This instance of ProtoRepeatedMessage.
        """
        # self._parent_proto.ClearField(self._repeated_field_name)
        del self._repeated_field_proto[:]
        return self

    def __str__(self):
        return str(self._repeated_field_proto)

    def __repr__(self):
        return str(self)


class ProtoBase(Generic[P]):
    """
    This class represents common routines between all sub-messages in API M.
    """

    def __init__(self, proto: P):
        self._proto: P = proto
        self._metadata: Metadata = Metadata(self._proto.metadata)

    def get_proto(self) -> P:
        """
        Return the backing protobuf for this class.
        :return: The backing protobuf for this class.
        """
        return self._proto

    def get_metadata(self) -> Metadata:
        """
        Returns the associated Metadata instance.
        :return: The associated Metadata instance.
        """
        return self._metadata

    def as_json(self) -> str:
        """
        Serializes and returns the backing protobuf as JSON.
        :return: The backing protobuf as JSON.
        """
        return MessageToJson(self._proto, True)

    def as_dict(self) -> Dict:
        """
        Serializes and returns the backing protobuf as a Python dictionary.
        :return: The backing protobuf as a Python dictionary.
        """
        return MessageToDict(self._proto, True)

    def as_bytes(self) -> bytes:
        """
        Serializes and returns the backing protobuf as bytes.
        :return: The backing protobuf as bytes.
        """
        return self._proto.SerializeToString()

    def as_compressed_bytes(self) -> bytes:
        """
        Serializes and returns the backing protobuf as compressed bytes.
        :return: The backing protobuf as compressed bytes.
        """
        data: bytes = self.as_bytes()
        return compress(data)

    def __str__(self):
        """
        Return the JSON representation of the backing protobuf,
        :return: The JSON representation of the backing protobuf.
        """
        return self.as_json()



from typing import List, Tuple, Optional
import enum
from math import modf
from dataclasses import dataclass, field

from dataclasses_json import dataclass_json
import pandas as pd
import numpy as np

from redvox.common import date_time_utils as dtu
from redvox.common.errors import RedVoxExceptions, RedVoxError
from redvox.api1000.wrapped_redvox_packet.sensors.audio import AudioCodec
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
from redvox.api1000.wrapped_redvox_packet.sensors.image import ImageCodec
from redvox.api1000.wrapped_redvox_packet.station_information import NetworkType, PowerState, CellServiceState

# default maximum number of points required to brute force calculating gap timestamps
DEFAULT_MAX_BRUTE_FORCE_GAP_TIMESTAMPS: int = 5000
# percent of packet duration/sample rate required for gap to be considered a whole unit
DEFAULT_GAP_UPPER_LIMIT: float = 0.8
# percent of packet duration/sample rate required for gap to be considered nothing
DEFAULT_GAP_LOWER_LIMIT: float = 0.02
# columns for audio dataframe
AUDIO_DF_COLUMNS = ["timestamps", "unaltered_timestamps", "microphone"]
# columns that cannot be interpolated
NON_INTERPOLATED_COLUMNS = ["compressed_audio", "image"]
# columns that are not numeric but can be interpolated
NON_NUMERIC_COLUMNS = ["location_provider", "image_codec", "audio_codec",
                       "network_type", "power_state", "cell_service"]


# noinspection Mypy,DuplicatedCode
class DataPointCreationMode(enum.Enum):
    """
    Type of data point to create
    """

    NAN: int = 0
    COPY: int = 1
    INTERPOLATE: int = 2

    @staticmethod
    def list_names() -> List[str]:
        return [n.name for n in DataPointCreationMode]


@dataclass_json()
@dataclass
class GapPadResult:
    """
    The result of filling gaps or padding a time series
    """
    result_df: Optional[pd.DataFrame] = None
    gaps: List[Tuple[float, float]] = field(default_factory=lambda: [])
    errors: RedVoxExceptions = field(default_factory=lambda: RedVoxExceptions("GapPadResult"))

    def add_error(self, error: str):
        """
        add an error to the result
        :param error: error message to add
        """
        self.errors.append(error)


def calc_evenly_sampled_timestamps(
        start: float, samples: int, sample_interval_micros: float
) -> np.array:
    """
    given a start time, calculates samples amount of evenly spaced timestamps at rate_hz

    :param start: float, start timestamp in microseconds
    :param samples: int, number of samples
    :param sample_interval_micros: float, sample interval in microseconds
    :return: np.array with number of samples timestamps, evenly spaced starting at start
    """
    return start + (np.arange(0, samples) * sample_interval_micros)


def check_gap_list(gaps: List[Tuple[float, float]], start_timestamp: float = None,
                   end_timestamp: float = None) -> List[Tuple[float, float]]:
    """
    removes any gaps where end time <= start time, consolidates overlapping gaps, and ensures that no gap
    starts or ends before start_timestamp and starts or ends after end_timestamp.  All timestamps are in
    microseconds since epoch UTC

    :param gaps: list of gaps to check
    :param start_timestamp: lowest possible timestamp for a gap to start at
    :param end_timestamp: lowest possible timestamp for a gap to end at
    :return: list of correct, valid gaps
    """
    return_gaps: List[Tuple[float, float]] = []
    for gap in gaps:
        if start_timestamp:
            gap = (np.max([start_timestamp, gap[0]]), np.max([start_timestamp, gap[1]]))
        if end_timestamp:
            gap = (np.min([end_timestamp, gap[0]]), np.min([end_timestamp, gap[1]]))
        if gap[0] < gap[1]:
            if len(return_gaps) < 1:
                return_gaps.append(gap)
            for a, r_g in enumerate(return_gaps):
                if (gap[0] < r_g[0] and gap[1] < r_g[0]) or (gap[0] > r_g[1] and gap[1] > r_g[1]):
                    return_gaps.append(gap)
                    break
                else:
                    if gap[0] < r_g[0] < gap[1]:
                        r_g = (gap[0], r_g[1])
                    if gap[0] < r_g[1] < gap[1]:
                        r_g = (r_g[0], gap[1])
                    return_gaps[a] = r_g
    return return_gaps


def pad_data(
        expected_start: float,
        expected_end: float,
        data_df: pd.DataFrame,
        sample_interval_micros: float,
) -> pd.DataFrame:
    """
    Pad the start and end of the dataframe with np.nan

    :param expected_start: timestamp indicating start time of the data to pad from
    :param expected_end: timestamp indicating end time of the data to pad from
    :param data_df: dataframe with timestamps as column "timestamps"
    :param sample_interval_micros: constant sample interval in microseconds
    :return: dataframe padded with np.nans in front and back to meet full size of expected start and end
    """
    # extract the necessary information to pad the data
    data_time_stamps = data_df["timestamps"].to_numpy()
    first_data_timestamp = data_time_stamps[0]
    last_data_timestamp = data_time_stamps[-1]
    result_df = data_df.copy()
    result_before_update_length = len(result_df) - 1
    # FRONT/END GAP FILL!  calculate the samples missing based on inputs
    if expected_start < first_data_timestamp:
        start_diff = first_data_timestamp - expected_start
        num_missing_samples = int(start_diff / sample_interval_micros)
        if num_missing_samples > 0:
            # add the gap data to the result dataframe
            result_df = add_dataless_timestamps_to_df(
                result_df,
                0,
                sample_interval_micros,
                num_missing_samples,
                True
            )
    if expected_end > last_data_timestamp:
        last_diff = expected_end - last_data_timestamp
        num_missing_samples = int(last_diff / sample_interval_micros)
        if num_missing_samples > 0:
            # add the gap data to the result dataframe
            result_df = add_dataless_timestamps_to_df(
                result_df,
                result_before_update_length,
                sample_interval_micros,
                num_missing_samples
            )
    return result_df.sort_values("timestamps", ignore_index=True)


def fill_gaps(
        data_df: pd.DataFrame,
        gaps: List[Tuple[float, float]],
        sample_interval_micros: float,
        copy: bool = False
) -> pd.DataFrame:
    """
    fills gaps in the dataframe with np.nan or interpolated values by interpolating timestamps based on the
    calculated sample interval

    :param data_df: dataframe with timestamps as column "timestamps"
    :param gaps: list of tuples of known non-inclusive start and end timestamps of the gaps
    :param sample_interval_micros: known sample interval of the data points
    :param copy: if True, copy the data points, otherwise interpolate from edges, default False
    :return: dataframe without gaps
    """
    # extract the necessary information to compute gap size and gap timestamps
    data_time_stamps = data_df["timestamps"].to_numpy()
    if len(data_time_stamps) > 1:
        result_df = data_df.copy()
        data_duration = data_time_stamps[-1] - data_time_stamps[0]
        expected_samples = (np.floor(data_duration / sample_interval_micros)
                            + (1 if data_duration % sample_interval_micros >=
                               sample_interval_micros * DEFAULT_GAP_UPPER_LIMIT else 0)) + 1
        if expected_samples > len(data_time_stamps):
            if copy:
                pcm = DataPointCreationMode["COPY"]
            else:
                pcm = DataPointCreationMode["NAN"]
            # make it safe to alter the gap values
            my_gaps = check_gap_list(gaps, data_time_stamps[0], data_time_stamps[-1])
            for gap in my_gaps:
                # if timestamps are around gaps, we have to update the values
                before_start = np.argwhere([t <= gap[0] for t in data_time_stamps])
                after_end = np.argwhere([t >= gap[1] for t in data_time_stamps])
                if len(before_start) > 0:
                    before_start = before_start[-1][0]
                    # sim = gap[0] - data_time_stamps[before_start]
                    # result_df = add_data_points_to_df(result_df, before_start, sim, point_creation_mode=pcm)
                    gap = (data_time_stamps[before_start], gap[1])
                else:
                    before_start = None
                if len(after_end) > 0:
                    after_end = after_end[0][0]
                    # sim = gap[1] - data_time_stamps[after_end]
                    gap = (gap[0], data_time_stamps[after_end])
                else:
                    after_end = None
                num_new_points = int((gap[1] - gap[0]) / sample_interval_micros) - 1
                if before_start is not None:
                    result_df = add_data_points_to_df(result_df, before_start, sample_interval_micros,
                                                      num_new_points, pcm)
                elif after_end is not None:
                    result_df = add_data_points_to_df(result_df, after_end, -sample_interval_micros,
                                                      num_new_points, pcm)
        return result_df.sort_values("timestamps", ignore_index=True)
    return data_df


def fill_audio_gaps(
        packet_data: List[Tuple[float, np.array]],
        sample_interval_micros: float,
        gap_upper_limit: float = DEFAULT_GAP_UPPER_LIMIT,
        gap_lower_limit: float = DEFAULT_GAP_LOWER_LIMIT
) -> GapPadResult:
    """
    fills gaps in the dataframe with np.nan by interpolating timestamps based on the expected sample interval
      * ignores gaps with duration less than or equal to packet length * gap_lower_limit
      * converts gaps with duration greater than or equal to packet length * gap_upper_limit into a multiple of
        packet length

    :param packet_data: list of tuples, each tuple containing two pieces of packet information:
        * packet_start_timestamps: float of packet start timestamp in microseconds
        * audio_data: array of data points
    :param sample_interval_micros: sample interval in microseconds
    :param gap_upper_limit: percentage of packet length required to confirm gap is at least 1 packet,
                            default DEFAULT_GAP_UPPER_LIMIT
    :param gap_lower_limit: percentage of packet length required to disregard gap, default DEFAULT_GAP_LOWER_LIMIT
    :return: dataframe without gaps and the list of timestamps of the non-inclusive start and end of the gaps
    """
    result_array = [[], [], []]
    last_data_timestamp: Optional[float] = None
    gaps = []
    for packet in packet_data:
        samples_in_packet = len(packet[1])
        start_ts = packet[0]
        packet_length = sample_interval_micros * samples_in_packet
        if last_data_timestamp:
            last_data_timestamp += sample_interval_micros
            # check if start_ts is close to the last timestamp in data_timestamps
            last_timestamp_diff = start_ts - last_data_timestamp
            if last_timestamp_diff > gap_lower_limit * packet_length:
                fractional_packet, num_packets = modf(last_timestamp_diff /
                                                      (samples_in_packet * sample_interval_micros))
                if fractional_packet >= gap_upper_limit:
                    num_samples = samples_in_packet * (num_packets + 1)
                else:
                    num_samples = np.max([np.floor((fractional_packet + num_packets) * samples_in_packet), 1])
                gap_ts = calc_evenly_sampled_timestamps(last_data_timestamp, num_samples, sample_interval_micros)
                gap_array = [gap_ts, np.full(len(gap_ts), np.nan)]
                start_ts = gap_ts[-1] + sample_interval_micros
                gaps.append((last_data_timestamp, start_ts))
                result_array[0].extend(gap_array[0])
                result_array[1].extend(gap_array[0])
                result_array[2].extend(gap_array[1])
            elif last_timestamp_diff < -gap_lower_limit * packet_length:
                result = GapPadResult()
                result.add_error(f"Packet start timestamp: {dtu.microseconds_to_seconds(start_ts)} "
                                 f"is before last timestamp of previous "
                                 f"packet: {dtu.microseconds_to_seconds(last_data_timestamp)}")
                return result
        estimated_ts = calc_evenly_sampled_timestamps(start_ts, samples_in_packet, sample_interval_micros)
        last_data_timestamp = estimated_ts[-1]
        result_array[0].extend(estimated_ts)
        result_array[1].extend(estimated_ts)
        result_array[2].extend(packet[1])
    return GapPadResult(pd.DataFrame(np.transpose(result_array), columns=AUDIO_DF_COLUMNS), gaps)


def add_data_points_to_df(dataframe: pd.DataFrame,
                          start_index: int,
                          sample_interval_micros: float,
                          num_samples_to_add: int = 1,
                          point_creation_mode: DataPointCreationMode = DataPointCreationMode.COPY,
                          ) -> pd.DataFrame:
    """
    adds data points to the end of the dataframe, starting from the index specified.
        Note:
            * dataframe must not be empty
            * start_index must be non-negative and less than the length of dataframe
            * num_samples_to_add must be greater than 0
            * points are added onto the end and the result is not sorted
        Options for point_creation_mode are:
            * NAN: default values and nans
            * COPY: copies of the start data point
            * INTERPOLATE: interpolated values between start data point and adjacent point

    :param dataframe: dataframe to add dataless timestamps to
    :param start_index: index of the dataframe to use as starting point for creating new values
    :param sample_interval_micros: sample interval in microseconds of the timestamps; use negative values to
                                    add points before the start_index
    :param num_samples_to_add: the number of timestamps to create, default 1
    :param point_creation_mode: the mode of point creation to use
    :return: updated dataframe with synthetic data points
    """
    if len(dataframe) > start_index and len(dataframe) > 0 and num_samples_to_add > 0:
        start_timestamp = dataframe["timestamps"].iloc[start_index]
        t = start_timestamp + np.arange(1, num_samples_to_add + 1) * sample_interval_micros
        # interpolate mode only uses the first created timestamp
        if point_creation_mode == DataPointCreationMode.COPY:
            empty_df = dataframe.iloc[start_index].copy()
            for column_index in dataframe.columns:
                if column_index in NON_INTERPOLATED_COLUMNS:
                    empty_df[column_index] = np.nan
            empty_df["timestamps"] = t[0]
        elif point_creation_mode == DataPointCreationMode.INTERPOLATE:
            start_point = dataframe.iloc[start_index]
            numeric_start = start_point[[col for col in dataframe.columns
                                         if col not in NON_INTERPOLATED_COLUMNS + NON_NUMERIC_COLUMNS]]
            non_numeric_start = start_point[[col for col in dataframe.columns if col in NON_NUMERIC_COLUMNS]]
            end_point = dataframe.iloc[start_index + (1 if sample_interval_micros > 0 else -1)]
            numeric_end = end_point[[col for col in dataframe.columns
                                     if col not in NON_INTERPOLATED_COLUMNS + NON_NUMERIC_COLUMNS]]
            non_numeric_end = end_point[[col for col in dataframe.columns if col in NON_NUMERIC_COLUMNS]]
            if np.abs(start_point["timestamps"] - t[0]) <= np.abs(end_point["timestamps"] - t[0]):
                non_numeric_diff = non_numeric_start
            else:
                non_numeric_diff = non_numeric_end
            numeric_diff = numeric_end - numeric_start
            numeric_diff = \
                (numeric_diff / numeric_diff["timestamps"]) * \
                (t - numeric_start) + numeric_start
            empty_df = pd.concat([numeric_diff, non_numeric_diff])
        else:
            empty_df = pd.DataFrame(np.full([num_samples_to_add, len(dataframe.columns)], np.nan),
                                    columns=dataframe.columns)
            for column_index in dataframe.columns:
                if column_index == "timestamps":
                    empty_df[column_index] = t
                elif column_index == "location_provider":
                    empty_df[column_index] = [LocationProvider["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "image_codec":
                    empty_df[column_index] = [ImageCodec["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "audio_codec":
                    empty_df[column_index] = [AudioCodec["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "network_type":
                    empty_df[column_index] = [NetworkType["UNKNOWN_NETWORK"].value for i in range(num_samples_to_add)]
                elif column_index == "power_state":
                    empty_df[column_index] = [PowerState["UNKNOWN_POWER_STATE"].value
                                              for i in range(num_samples_to_add)]
                elif column_index == "cell_service":
                    empty_df[column_index] = [CellServiceState["UNKNOWN"].value for i in range(num_samples_to_add)]
        dataframe = dataframe.append(empty_df, ignore_index=True)

    return dataframe


def add_dataless_timestamps_to_df(dataframe: pd.DataFrame,
                                  start_index: int,
                                  sample_interval_micros: float,
                                  num_samples_to_add: int,
                                  add_to_start: bool = False,
                                  copy: bool = True,
                                  ) -> pd.DataFrame:
    """
    adds dataless timestamps directly to a dataframe that already contains data
      Note:
        * dataframe must not be empty
        * start_index must be non-negative and less than the length of dataframe
        * num_samples_to_add must be greater than 0
        * the points are added onto the end and the result is not sorted

    :param dataframe: dataframe to add dataless timestamps to
    :param start_index: index of the dataframe to use as starting point for creating new values
    :param sample_interval_micros: sample interval in microseconds of the timestamps
    :param num_samples_to_add: the number of timestamps to create
    :param add_to_start: if True, subtracts sample_interval_micros from start_timestamp, default False
    :param copy: if True, copy the value of the start point when creating new points, default True
    :return: updated dataframe with synthetic data points
    """
    if len(dataframe) > start_index and len(dataframe) > 0 and num_samples_to_add > 0:
        start_timestamp = dataframe["timestamps"].iloc[start_index]
        dataframe = dataframe.append(
            create_dataless_timestamps_df(start_timestamp, sample_interval_micros,
                                          dataframe.columns, num_samples_to_add, add_to_start),
            ignore_index=True)
    return dataframe


def create_dataless_timestamps_df(
        start_timestamp: float,
        sample_interval_micros: float,
        columns: pd.Index,
        num_samples_to_add: int,
        add_to_start: bool = False,
) -> pd.DataFrame:
    """
    Creates an empty dataframe with num_samples_to_add timestamps, using columns as the columns
    the first timestamp created is 1 sample_interval_s from the start_timestamp

    :param start_timestamp: timestamp in microseconds since epoch UTC to start calculating other timestamps from
    :param sample_interval_micros: fixed sample interval in microseconds since epoch UTC
    :param columns: dataframe the non-timestamp columns of the dataframe
    :param num_samples_to_add: the number of timestamps to create
    :param add_to_start: if True, subtracts sample_interval_s from start_timestamp, default False
    :return: dataframe with timestamps and no data
    """
    empty_df = pd.DataFrame(np.full([num_samples_to_add, len(columns)], np.nan), columns=columns)
    enum_samples = {
        "location_provider": LocationProvider["UNKNOWN"].value,
        "image_codec": ImageCodec["UNKNOWN"].value,
        "audio_codec": AudioCodec["UNKNOWN"].value,
        "network_type": NetworkType["UNKNOWN_NETWORK"].value,
        "power_state": PowerState["UNKNOWN_POWER_STATE"].value,
        "cell_service": CellServiceState["UNKNOWN"].value
    }
    if num_samples_to_add > 0:
        if add_to_start:
            sample_interval_micros = -sample_interval_micros
        t = start_timestamp + np.arange(1, num_samples_to_add + 1) * sample_interval_micros
        for column_index in columns:
            if column_index == "timestamps":
                empty_df[column_index] = t
            elif column_index in enum_samples.keys():
                empty_df[column_index] = [enum_samples[column_index] for i in range(num_samples_to_add)]
            # elif column_index == "location_provider":
            #     empty_df[column_index] = [LocationProvider.UNKNOWN for i in range(num_samples_to_add)]
            # elif column_index == "image_codec":
            #     empty_df[column_index] = [ImageCodec.UNKNOWN for i in range(num_samples_to_add)]
            # elif column_index == "audio_codec":
            #     empty_df[column_index] = [AudioCodec.UNKNOWN for i in range(num_samples_to_add)]
            # elif column_index == "network_type":
            #     empty_df[column_index] = [NetworkType.UNKNOWN_NETWORK for i in range(num_samples_to_add)]
            # elif column_index == "power_state":
            #     empty_df[column_index] = [PowerState.UNKNOWN_POWER_STATE for i in range(num_samples_to_add)]
            # elif column_index == "cell_service":
            #     empty_df[column_index] = [CellServiceState.UNKNOWN for i in range(num_samples_to_add)]
    return empty_df



"""
This module creates specific time-bounded segments of data for users
combines the base data files into a single composite object based on the user parameters
"""
from pathlib import Path
from typing import Optional, Set, List, Dict, Iterable
from datetime import timedelta

import multiprocessing
import multiprocessing.pool
import pickle
import numpy as np

import redvox
from redvox.common import date_time_utils as dtu
from redvox.common import io
from redvox.common import data_window_io as dw_io
from redvox.common.parallel_utils import maybe_parallel_map
from redvox.common.station import Station
from redvox.common.sensor_data import SensorType, SensorData
from redvox.common.api_reader import ApiReader
from redvox.common.data_window_configuration import DataWindowConfig
from redvox.common import gap_and_pad_utils as gpu
from redvox.common.errors import RedVoxExceptions

DEFAULT_START_BUFFER_TD: timedelta = timedelta(minutes=2.0)  # default padding to start time of data
DEFAULT_END_BUFFER_TD: timedelta = timedelta(minutes=2.0)  # default padding to end time of data
# minimum default length of time in seconds for data to be off by to be considered suspicious
DATA_DROP_DURATION_S: float = 0.2


class DataWindow:
    """
    Holds the data for a given time window; adds interpolated timestamps to fill gaps and pad start and end values

    Properties:
        input_directory: string, directory that contains the files to read data from.  REQUIRED
        structured_layout: bool, if True, the input_directory contains specially named and organized
                            directories of data.  Default True
        station_ids: optional set of strings, representing the station ids to filter on.
                        If empty or None, get any ids found in the input directory.  Default None
        extensions: optional set of strings, representing file extensions to filter on.
                        If None, gets as much data as it can in the input directory.  Default None
        api_versions: optional set of ApiVersions, representing api versions to filter on.
                        If None, get as much data as it can in the input directory.  Default None
        start_datetime: optional datetime, start datetime of the window.
                        If None, uses the first timestamp of the filtered data.  Default None
        end_datetime: optional datetime, non-inclusive end datetime of the window.
                        If None, uses the last timestamp of the filtered data + 1.  Default None
        start_buffer_td: timedelta, the amount of time to include before the start_datetime when filtering data.
                            Negative values are converted to 0.  Default DEFAULT_START_BUFFER_TD (2 minutes)
        end_buffer_td: timedelta, the amount of time to include after the end_datetime when filtering data.
                            Negative values are converted to 0.  Default DEFAULT_END_BUFFER_TD (2 minutes)
        drop_time_s: float, the minimum amount of seconds between data files that would indicate a gap.
                     Negative values are converted to default value.  Default DATA_DROP_DURATION_S (0.2 seconds)
        apply_correction: bool, if True, update the timestamps in the data based on best station offset.  Default True
        use_model_correction: bool, if True, use the offset model's correction functions, otherwise use the best
                                offset.  Default True
        copy_edge_points: enumeration of DataPointCreationMode.  Determines how new points are created.
                            Valid values are NAN, COPY, and INTERPOLATE.  Default COPY
        debug: bool, if True, outputs additional information during initialization. Default False
        errors: DataWindowExceptions, class containing a list of all errors encountered by the data window.
        stations: list of Stations, the results of reading the data from input_directory
        sdk_version: str, the version of the Redvox SDK used to create the data window
    """
    def __init__(
            self,
            input_dir: str,
            structured_layout: bool = True,
            start_datetime: Optional[dtu.datetime] = None,
            end_datetime: Optional[dtu.datetime] = None,
            start_buffer_td: timedelta = DEFAULT_START_BUFFER_TD,
            end_buffer_td: timedelta = DEFAULT_END_BUFFER_TD,
            drop_time_s: float = DATA_DROP_DURATION_S,
            station_ids: Optional[Iterable[str]] = None,
            extensions: Optional[Set[str]] = None,
            api_versions: Optional[Set[io.ApiVersion]] = None,
            apply_correction: bool = True,
            copy_edge_points: gpu.DataPointCreationMode = gpu.DataPointCreationMode.COPY,
            debug: bool = False,
            use_model_correction: bool = True,
    ):
        """
        Initialize the DataWindow

        :param input_dir: directory that contains the files to read data from.  REQUIRED
        :param structured_layout: if True, the input_directory contains specially named and organized directories of
                                    data.  Default True
        :param start_datetime: optional start datetime of the window.  If None, uses the first timestamp of the
                                filtered data.  Default None
        :param end_datetime: optional non-inclusive end datetime of the window.  If None, uses the last timestamp of
                                the filtered data + 1.  Default None
        :param start_buffer_td: the amount of time to include before the start_datetime when filtering data.
                                Negative values are converted to 0.  Default 2 minutes
        :param end_buffer_td: the amount of time to include after the end_datetime when filtering data.
                                Negative values are converted to 0.  Default 2 minutes
        :param station_ids: optional station ids to filter on. If empty or None, get any ids found in the input
                            directory.  Default None
        :param drop_time_s: the minimum amount of seconds between data files that would indicate a gap.
                            Negative values are converted to default value.  Default 0.2 seconds
        :param extensions: optional set of file extensions to filter on.  If None, gets as much data as it can in the
                            input directory.  Default None
        :param api_versions: optional set of api versions to filter on.  If None, get as much data as it can in
                                the input directory.  Default None
        :param apply_correction: if True, update the timestamps in the data based on best station offset.  Default True
        :param copy_edge_points: Determines how new points are created. Valid values are DataPointCreationMode.NAN,
                                    DataPointCreationMode.COPY, and DataPointCreationMode.INTERPOLATE.  Default COPY
        :param use_model_correction: if True, use the offset model's correction functions, otherwise use the best
                                        offset.  Default True
        :param debug: if True, outputs additional information during initialization. Default False
        """
        self.errors = RedVoxExceptions("DataWindow")
        self.input_directory: str = input_dir
        self.structured_layout: bool = structured_layout
        self.start_datetime: Optional[dtu.datetime] = start_datetime
        self.end_datetime: Optional[dtu.datetime] = end_datetime
        self.start_buffer_td: timedelta = start_buffer_td if start_buffer_td > timedelta(seconds=0) \
            else timedelta(seconds=0)
        self.end_buffer_td: timedelta = end_buffer_td if end_buffer_td > timedelta(seconds=0) else timedelta(seconds=0)
        self.drop_time_s: float = drop_time_s if drop_time_s > 0 else DATA_DROP_DURATION_S
        self.station_ids: Optional[Set[str]]
        if station_ids:
            self.station_ids = set(station_ids)
        else:
            self.station_ids = None
        self.extensions: Optional[Set[str]] = extensions
        self.api_versions: Optional[Set[io.ApiVersion]] = api_versions
        self.apply_correction: bool = apply_correction
        self.copy_edge_points = copy_edge_points
        self.use_model_correction = use_model_correction
        self.sdk_version: str = redvox.VERSION
        self.debug: bool = debug
        self.stations: List[Station] = []
        if start_datetime and end_datetime and (end_datetime <= start_datetime):
            self.errors.append("DataWindow will not work when end datetime is before or equal to start datetime.\n"
                               f"Your times: {end_datetime} <= {start_datetime}")
        else:
            self.create_data_window()
        if debug:
            self.print_errors()

    @staticmethod
    def from_config_file(file: str) -> "DataWindow":
        """
        Loads a configuration file to create the DataWindow

        :param file: full path to config file
        :return: a data window
        """
        return DataWindow.from_config(DataWindowConfig.from_path(file))

    @staticmethod
    def from_config(config: DataWindowConfig) -> "DataWindow":
        """
        Loads a configuration to create the DataWindow

        :param config: DataWindow configuration object
        :return: a data window
        """
        if config.start_year:
            start_time = dtu.datetime(
                year=config.start_year,
                month=config.start_month,
                day=config.start_day,
                hour=config.start_hour,
                minute=config.start_minute,
                second=config.start_second,
            )
        else:
            start_time = None
        if config.end_year:
            end_time = dtu.datetime(
                year=config.end_year,
                month=config.end_month,
                day=config.end_day,
                hour=config.end_hour,
                minute=config.end_minute,
                second=config.end_second,
            )
        else:
            end_time = None
        if config.api_versions:
            api_versions = set([io.ApiVersion.from_str(v) for v in config.api_versions])
        else:
            api_versions = None
        if config.extensions:
            extensions = set(config.extensions)
        else:
            extensions = None
        if config.station_ids:
            station_ids = set(config.station_ids)
        else:
            station_ids = None
        if config.edge_points_mode not in gpu.DataPointCreationMode.list_names():
            config.edge_points_mode = "COPY"
        return DataWindow(
            config.input_directory,
            config.structured_layout,
            start_time,
            end_time,
            dtu.timedelta(seconds=config.start_padding_seconds),
            dtu.timedelta(seconds=config.end_padding_seconds),
            config.drop_time_seconds,
            station_ids,
            extensions,
            api_versions,
            config.apply_correction,
            gpu.DataPointCreationMode[config.edge_points_mode],
            config.debug,
            config.use_model_correction,
        )

    @staticmethod
    def deserialize(path: str) -> "DataWindow":
        """
        Decompresses and deserializes a DataWindow written to disk.

        :param path: Path to the serialized and compressed data window.
        :return: An instance of a DataWindow.
        """
        return dw_io.deserialize_data_window(path)

    def serialize(self, base_dir: str = ".", file_name: Optional[str] = None, compression_factor: int = 4) -> Path:
        """
        Serializes and compresses this DataWindow to a file.

        :param base_dir: The base directory to write the serialized file to (default=.).
        :param file_name: The optional file name. If None, a default filename with the following format is used:
                          [start_ts]_[end_ts]_[num_stations].pkl.lz4
        :param compression_factor: A value between 1 and 12. Higher values provide better compression, but take
        longer. (default=4).
        :return: The path to the written file.
        """
        return dw_io.serialize_data_window(self, base_dir, file_name, compression_factor)

    def to_json_file(self, base_dir: str = ".", file_name: Optional[str] = None,
                     compression_format: str = "lz4") -> Path:
        """
        Converts the data window metadata into a JSON file and compresses the data window and writes it to disk.

        :param base_dir: base directory to write the json file to.  Default . (local directory)
        :param file_name: the optional file name.  Do not include a file extension.
                            If None, a default file name is created using this format:
                            [start_ts]_[end_ts]_[num_stations].json
        :param compression_format: the type of compression to use on the data window object.  default lz4
        :return: The path to the written file
        """
        return dw_io.data_window_to_json_file(self, base_dir, file_name, compression_format)

    def to_json(self, compressed_file_base_dir: str = ".", compressed_file_name: Optional[str] = None,
                compression_format: str = "lz4") -> str:
        """
        Converts the data window metadata into a JSON string, then compresses the data window and writes it to disk.

        :param compressed_file_base_dir: base directory to write the json file to.  Default . (local directory)
        :param compressed_file_name: the optional file name.  Do not include a file extension.
                                        If None, a default file name is created using this format:
                                        [start_ts]_[end_ts]_[num_stations].[compression_format]
        :param compression_format: the type of compression to use on the data window object.  default lz4
        :return: The json string
        """
        return dw_io.data_window_to_json(self, compressed_file_base_dir, compressed_file_name, compression_format)

    @staticmethod
    def from_json_file(base_dir: str, file_name: str,
                       dw_base_dir: Optional[str] = None,
                       start_dt: Optional[dtu.datetime] = None,
                       end_dt: Optional[dtu.datetime] = None,
                       station_ids: Optional[Iterable[str]] = None) -> Optional["DataWindow"]:
        """
        Reads a JSON file and checks if:
            * The requested times are within the JSON file's times
            * The requested stations are a subset of the JSON file's stations

        :param base_dir: the base directory containing the json file
        :param file_name: the file name of the json file.  Do not include extensions
        :param dw_base_dir: optional directory containing the compressed data window file.
                            If not given, assume in subdirectory "dw".  default None
        :param start_dt: the start datetime to check against.  if not given, assumes True.  default None
        :param end_dt: the end datetime to check against.  if not given, assumes True.  default None
        :param station_ids: the station ids to check against.  if not given, assumes True.  default None
        :return: the data window if it suffices, otherwise None
        """
        if not dw_base_dir:
            dw_base_dir = Path(base_dir).joinpath("dw")
        file_name += ".json"
        return DataWindow.from_json_dict(
            dw_io.json_file_to_data_window(base_dir, file_name), dw_base_dir, start_dt, end_dt, station_ids)

    @staticmethod
    def from_json(json_str: str, dw_base_dir: str,
                  start_dt: Optional[dtu.datetime] = None,
                  end_dt: Optional[dtu.datetime] = None,
                  station_ids: Optional[Iterable[str]] = None) -> Optional["DataWindow"]:
        """
        Reads a JSON string and checks if:
            * The requested times are within the JSON file's times
            * The requested stations are a subset of the JSON file's stations

        :param json_str: the JSON to read
        :param dw_base_dir: directory containing the compressed data window file
        :param start_dt: the start datetime to check against.  if not given, assumes True.  default None
        :param end_dt: the end datetime to check against.  if not given, assumes True.  default None
        :param station_ids: the station ids to check against.  if not given, assumes True.  default None
        :return: the data window if it suffices, otherwise None
        """
        return DataWindow.from_json_dict(dw_io.json_to_data_window(json_str), dw_base_dir,
                                         start_dt, end_dt, station_ids)

    @staticmethod
    def from_json_dict(json_dict: Dict,
                       dw_base_dir: str,
                       start_dt: Optional[dtu.datetime] = None,
                       end_dt: Optional[dtu.datetime] = None,
                       station_ids: Optional[Iterable[str]] = None) -> Optional["DataWindow"]:
        """
        Reads a JSON string and checks if:
            * The requested times are within the JSON file's times
            * The requested stations are a subset of the JSON file's stations

        :param json_dict: the dictionary to read
        :param dw_base_dir: base directory for the compressed data window file
        :param start_dt: optional start datetime to check against.  if not given, assumes True.  default None
        :param end_dt: optional end datetime to check against.  if not given, assumes True.  default None
        :param station_ids: optional station ids to check against.  if not given, assumes True.  default None
        :return: the data window if it suffices, otherwise None
        """
        if start_dt and json_dict["start_datetime"] >= dtu.datetime_to_epoch_microseconds_utc(start_dt):
            return None
        if end_dt and json_dict["end_datetime"] < dtu.datetime_to_epoch_microseconds_utc(end_dt):
            return None
        if station_ids and not all(a in json_dict["station_ids"] for a in station_ids):
            return None
        comp_dw_path = str(Path(dw_base_dir).joinpath(json_dict["file_name"]))
        if json_dict["compression_format"] == "lz4":
            return DataWindow.deserialize(comp_dw_path + ".pkl.lz4")
        else:
            with open(comp_dw_path + ".pkl", 'rb') as fp:
                return pickle.load(fp)

    def get_station(self, station_id: str, station_uuid: Optional[str] = None,
                    start_timestamp: Optional[float] = None) -> Optional[List[Station]]:
        """
        Get stations from the data window.  Must give at least the station's id.  Other parameters may be None,
        which means the value will be ignored when searching.  Results will match all non-None parameters given.

        :param station_id: station id
        :param station_uuid: station uuid, default None
        :param start_timestamp: station start timestamp in microseconds since UTC epoch, default None
        :return: A list of valid stations or None if the station cannot be found
        """
        result = [s for s in self.stations if s.get_key().check_key(station_id, station_uuid, start_timestamp)]
        if len(result) > 0:
            return result
        if self.debug:
            self.errors.append(f"Attempted to get station {station_id}, "
                               f"but that station is not in this data window!")
        return None

    def _add_sensor_to_window(self, station: Station):
        self.errors.extend_error(station.errors)
        # set the window start and end if they were specified, otherwise use the bounds of the data
        self.create_window_in_sensors(station, self.start_datetime, self.end_datetime)

    def create_data_window(self, pool: Optional[multiprocessing.pool.Pool] = None):
        """
        updates the data window to contain only the data within the window parameters
        stations without audio or any data outside the window are removed
        """
        # Let's create and manage a single pool of workers that we can utilize throughout
        # the instantiation of the data window.
        _pool: multiprocessing.pool.Pool = multiprocessing.Pool() if pool is None else pool

        r_f = io.ReadFilter()
        if self.start_datetime:
            r_f.with_start_dt(self.start_datetime)
        if self.end_datetime:
            r_f.with_end_dt(self.end_datetime)
        if self.station_ids:
            r_f.with_station_ids(self.station_ids)
        if self.extensions:
            r_f.with_extensions(self.extensions)
        else:
            self.extensions = r_f.extensions
        if self.api_versions:
            r_f.with_api_versions(self.api_versions)
        else:
            self.api_versions = r_f.api_versions
        r_f.with_start_dt_buf(self.start_buffer_td)
        r_f.with_end_dt_buf(self.end_buffer_td)

        # get the data to convert into a window
        a_r = ApiReader(self.input_directory, self.structured_layout, r_f, pool=_pool)

        self.errors.extend_error(a_r.errors)

        if not self.station_ids:
            self.station_ids = a_r.index_summary.station_ids()
        # Parallel update
        # Apply timing correction in parallel by station
        sts = a_r.get_stations()
        if not self.use_model_correction:
            for tss in sts:
                tss.use_model_correction = self.use_model_correction
        if self.apply_correction:
            for st in maybe_parallel_map(_pool, Station.update_timestamps,
                                         iter(sts), chunk_size=1):
                self._add_sensor_to_window(st)
        else:
            [self._add_sensor_to_window(s) for s in sts]
        # check for stations without data
        self._check_for_audio()
        self._check_valid_ids()

        # update remaining data window values if they're still default
        if not self.start_datetime and len(self.stations) > 0:
            self.start_datetime = dtu.datetime_from_epoch_microseconds_utc(
                np.min([t.first_data_timestamp for t in self.stations]))
        # end_datetime is non-inclusive, so it must be greater than our latest timestamp
        if not self.end_datetime and len(self.stations) > 0:
            self.end_datetime = dtu.datetime_from_epoch_microseconds_utc(
                np.max([t.last_data_timestamp for t in self.stations]) + 1)

        # If the pool was created by this function, then it needs to managed by this function.
        if pool is None:
            _pool.close()

    def _check_for_audio(self):
        """
        removes any station and station id without audio data from the data window
        """
        remove = []
        for s in self.stations:
            if not s.has_audio_sensor():
                remove.append(s.id)
        if len(remove) > 0:
            self.stations = [s for s in self.stations if s.id not in remove]
            self.station_ids = [s for s in self.station_ids if s not in remove]

    def _check_valid_ids(self):
        """
        if there are stations, searches the station_ids for any ids not in the data collected
        and creates an error message for each id requested but has no data
        if there are no stations, creates a single error message declaring no data found
        """
        if len(self.stations) < 1:
            if len(self.station_ids) > 1:
                add_ids = f"for all stations {self.station_ids} "
            else:
                add_ids = ""
            self.errors.append(f"No data matching criteria {add_ids}in {self.input_directory}"
                               f"\nPlease adjust parameters of DataWindow")
        elif len(self.station_ids) > 1:
            for ids in self.station_ids:
                if ids not in [i.id for i in self.stations] and self.debug:
                    self.errors.append(
                        f"Requested {ids} but there is no data to read for that station"
                    )

    def create_window_in_sensors(
            self, station: Station, start_datetime: Optional[dtu.datetime] = None,
            end_datetime: Optional[dtu.datetime] = None
    ):
        """
        truncate the sensors in the station to only contain data from start_date_timestamp to end_date_timestamp
        if the start and/or end are not specified, keeps all audio data that fits and uses it
        to truncate the other sensors.
        returns nothing, updates the station in place

        :param station: station object to truncate sensors of
        :param start_datetime: datetime of start of window, default None
        :param end_datetime: datetime of end of window, default None
        """
        if start_datetime:
            start_datetime = dtu.datetime_to_epoch_microseconds_utc(start_datetime)
        else:
            start_datetime = 0
        if end_datetime:
            end_datetime = dtu.datetime_to_epoch_microseconds_utc(end_datetime)
        else:
            end_datetime = dtu.datetime_to_epoch_microseconds_utc(dtu.datetime.max)
        self.process_sensor(station.audio_sensor(), station.id, start_datetime, end_datetime)
        for sensor in [s for s in station.data if s.type != SensorType.AUDIO]:
            self.process_sensor(sensor, station.id, station.audio_sensor().first_data_timestamp(),
                                station.audio_sensor().last_data_timestamp())
        # recalculate metadata
        station.first_data_timestamp = station.audio_sensor().first_data_timestamp()
        station.last_data_timestamp = station.audio_sensor().data_timestamps()[-1]
        station.packet_metadata = [meta for meta in station.packet_metadata
                                   if meta.packet_start_mach_timestamp < station.last_data_timestamp and
                                   meta.packet_end_mach_timestamp >= station.first_data_timestamp]
        self.stations.append(station)

    def process_sensor(self, sensor: SensorData, station_id: str, start_date_timestamp: float,
                       end_date_timestamp: float):
        """
        process a non audio sensor to fit within the data window.  Updates sensor in place, returns nothing.

        :param sensor: sensor to process
        :param station_id: station id
        :param start_date_timestamp: start of data window according to data
        :param end_date_timestamp: end of data window according to data
        """
        if sensor.num_samples() > 0:
            # get only the timestamps between the start and end timestamps
            before_start = np.where(sensor.data_timestamps() < start_date_timestamp)[0]
            after_end = np.where(end_date_timestamp <= sensor.data_timestamps())[0]
            # start_index is inclusive of window start
            if len(before_start) > 0:
                last_before_start = before_start[-1]
                start_index = last_before_start + 1
            else:
                last_before_start = None
                start_index = 0
            # end_index is non-inclusive of window end
            if len(after_end) > 0:
                first_after_end = after_end[0]
                end_index = first_after_end
            else:
                first_after_end = None
                end_index = sensor.num_samples()
            # check if all the samples have been cut off
            is_audio = sensor.type == SensorType.AUDIO
            if end_index <= start_index:
                if is_audio:
                    self.errors.append(f"Data window for {station_id} "
                                       f"Audio sensor has truncated all data points")
                elif last_before_start is not None and first_after_end is None:
                    sensor.data_df = sensor.data_df.iloc[[last_before_start]]
                    sensor.data_df["timestamps"] = [start_date_timestamp]
                elif last_before_start is None and first_after_end is not None:
                    sensor.data_df = sensor.data_df.iloc[[first_after_end]]
                    sensor.data_df["timestamps"] = [end_date_timestamp]
                elif last_before_start is not None and first_after_end is not None:
                    sensor.data_df = sensor.interpolate(start_date_timestamp, last_before_start, 1,
                                                        self.copy_edge_points == gpu.DataPointCreationMode.COPY
                                                        ).to_frame().T
                else:
                    self.errors.append(
                        f"Data window for {station_id} {sensor.type.name} "
                        f"sensor has truncated all data points"
                    )
            else:
                sensor.data_df = sensor.data_df.iloc[start_index:end_index].reset_index(
                    drop=True
                )
                # if sensor is audio or location, we want nan'd edge points
                if sensor.type in [SensorType.LOCATION, SensorType.AUDIO]:
                    new_point_mode = gpu.DataPointCreationMode["NAN"]
                else:
                    new_point_mode = self.copy_edge_points
                # add in the data points at the edges of the window if there are defined start and/or end times
                if not is_audio:
                    end_sample_interval = end_date_timestamp - sensor.last_data_timestamp()
                    end_samples_to_add = 1
                    start_sample_interval = start_date_timestamp - sensor.first_data_timestamp()
                    start_samples_to_add = 1
                else:
                    end_sample_interval = dtu.seconds_to_microseconds(sensor.sample_interval_s)
                    start_sample_interval = -end_sample_interval
                    if self.end_datetime:
                        end_samples_to_add = int((dtu.datetime_to_epoch_microseconds_utc(self.end_datetime)
                                                  - sensor.last_data_timestamp()) / end_sample_interval)
                    else:
                        end_samples_to_add = 0
                    if self.start_datetime:
                        start_samples_to_add = int((sensor.first_data_timestamp() -
                                                    dtu.datetime_to_epoch_microseconds_utc(self.start_datetime))
                                                   / end_sample_interval)
                    else:
                        start_samples_to_add = 0
                # add to end
                sensor.data_df = gpu.add_data_points_to_df(dataframe=sensor.data_df,
                                                           start_index=sensor.num_samples() - 1,
                                                           sample_interval_micros=end_sample_interval,
                                                           num_samples_to_add=end_samples_to_add,
                                                           point_creation_mode=new_point_mode)
                # add to begin
                sensor.data_df = gpu.add_data_points_to_df(dataframe=sensor.data_df, start_index=0,
                                                           sample_interval_micros=start_sample_interval,
                                                           num_samples_to_add=start_samples_to_add,
                                                           point_creation_mode=new_point_mode)
                sensor.data_df.sort_values("timestamps", inplace=True, ignore_index=True)
        else:
            self.errors.append(f"Data window for {station_id} {sensor.type.name} "
                               f"sensor has no data points!")

    def print_errors(self):
        """
        prints errors to screen
        """
        self.errors.print()



"""
This module provides functionality for working with API M image sensors.
"""

import enum
import os.path
from typing import List, Optional

import redvox.api1000.common.common as common
from redvox.api1000.common.decorators import wrap_enum
import redvox.api1000.common.generic
from redvox.api1000.common.typing import check_type
from redvox.api1000.errors import ApiMImageChannelError
import redvox.api1000.proto.redvox_api_m_pb2 as redvox_api_m_pb2


# noinspection Mypy
@wrap_enum(redvox_api_m_pb2.RedvoxPacketM.Sensors.Image.ImageCodec)
class ImageCodec(enum.Enum):
    """
    Image Codecs for image data
    """

    UNKNOWN: int = 0
    PNG: int = 1
    JPG: int = 2
    BMP: int = 3


class Image(
    redvox.api1000.common.generic.ProtoBase[
        redvox_api_m_pb2.RedvoxPacketM.Sensors.Image
    ]
):
    """
    This class encapsulates metadata and data associated with the image sensor.
    """

    def __init__(self, proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.Image):
        super().__init__(proto)
        self._timestamps: common.TimingPayload = common.TimingPayload(proto.timestamps)

    @staticmethod
    def new() -> "Image":
        """
        :return: A new, empty Image sensor instance
        """
        return Image(redvox_api_m_pb2.RedvoxPacketM.Sensors.Image())

    def get_image_codec(self) -> ImageCodec:
        """
        :return: Returns the codec that was used to store the images.
        """
        # noinspection Mypy
        # pylint: disable=E1101
        return ImageCodec.from_proto(self._proto.image_codec)

    def get_file_ext(self) -> str:
        """
        :return: The file extension for the given codec.
        """
        return self.get_image_codec().name.lower()

    def set_image_codec(self, codec: ImageCodec) -> "Image":
        """
        Sets the codec used to store the images.
        :param codec: Codec to set.
        :return: A modified instance of self
        """
        check_type(codec, [ImageCodec])
        # noinspection Mypy
        self._proto.image_codec = codec.into_proto()
        return self

    def get_sensor_description(self) -> str:
        """
        :return: The image sensor description.
        """
        return self._proto.sensor_description

    def set_sensor_description(self, sensor_description: str) -> "Image":
        """
        Sets the image sensor description.
        :param sensor_description: Description to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sensor_description, [str])
        self._proto.sensor_description = sensor_description
        return self

    def get_timestamps(self) -> common.TimingPayload:
        """
        :return: Timestamp payload which contains a timestamp per image
        """
        return self._timestamps

    def set_timestamps(self, timestamps: common.TimingPayload) -> "Image":
        """
        Sets the timestamps.
        :param timestamps: Timestamps to set.
        :return: A modified instance of self.
        """
        check_type(timestamps, [common.TimingPayload])
        self.get_proto().timestamps.CopyFrom(timestamps.get_proto())
        self._timestamps = common.TimingPayload(self.get_proto().timestamps)
        return self

    def get_samples(self) -> List[bytes]:
        """
        :return: Returns each image as collection of bytes.
        """
        return list(self.get_proto().samples)

    def set_samples(self, images: List[bytes]) -> "Image":
        """
        Set images.
        :param images: List of bytes objects representing each image.
        :return: A modified instance of self
        """
        # check_type(images, [List[bytes]])
        self._proto.samples[:] = images
        return self

    def append_value(self, image: bytes) -> "Image":
        """
        Appends a single image to this sensors list of images.
        :param image: Image to append as serialized bytes.
        :return: A modified instance of self
        """
        check_type(image, [bytes])
        self._proto.samples.append(image)
        return self

    def append_values(self, images: List[bytes]) -> "Image":
        """
        Appends multiple images to this sensor's image payload.
        :param images: Images to append as a list of bytes objects.
        :return: A modified instance of self
        """
        check_type(images, [List])
        self._proto.samples.extend(list(images))
        return self

    def clear_values(self) -> "Image":
        """
        Clears all images from this sensor
        :return: A modified instance of self
        """
        self._proto.samples[:] = []
        return self

    def get_num_images(self) -> int:
        """
        :return: The total number of images stored in this packet
        """
        return len(self.get_samples())

    def write_image(
        self, index: int, base_dir: str = ".", out_file: Optional[str] = None
    ) -> str:
        """
        Writes an image to disk.
        :param base_dir: Base directory to write image to (default: ".")
        :param out_file: Optional file name (sans extension) (default: timestamp of image)
        :param index: Index of image to be written.
        :return: Path of written file.
        """
        if index < 0 or index >= self.get_num_images():
            raise ApiMImageChannelError(
                f"Index={index} must be > 0 and < {self.get_num_images()}"
            )

        ext: str = self.get_file_ext()
        base_name: str = (
            str(int(self._timestamps.get_timestamps()[index]))
            if out_file is None
            else out_file
        )
        file_name: str = f"{base_name}.{ext}"
        file_path: str = os.path.join(base_dir, file_name)

        with open(file_path, "wb") as image_out:
            img_bytes: bytes = self.get_samples()[index]
            image_out.write(img_bytes)

        return file_path

    def write_images(
        self, indices: Optional[List[int]] = None, base_dir: str = "."
    ) -> List[str]:
        """
        Write multiple images to disk.
        :param base_dir: Base directory to write images to (default: ".").
        :param indices: Optional list of indices for images to write. If this is None, all images will be written.
        :return: List of written file paths.
        """

        indices = (
            indices if indices is not None else list(range(0, self.get_num_images()))
        )
        return list(map(lambda i: self.write_image(i, base_dir), indices))

    def gallery(self):
        """
        Opens a simple image gallery for viewing images in packet
        """
        try:
            import redvox.api1000.gui.image_viewer as image_viewer
            image_viewer.start_gui(self)
        except ImportError:
            import warnings
            warnings.warn("GUI dependencies are not installed. Install the 'GUI' extra to enable this functionality.")


def validate_image(image_sensor: Image) -> List[str]:
    """
    Validates the image sensor.
    :param image_sensor: Image sensor to validate.
    :return: A list of validation errors.
    """
    errors_list = common.validate_timing_payload(image_sensor.get_timestamps())
    if image_sensor.get_image_codec() not in ImageCodec.__members__.values():
        errors_list.append("Image sensor image codec is unknown")
    return errors_list



"""
Defines generic sensor data and data for API-independent analysis
all timestamps are integers in microseconds unless otherwise stated
"""
import enum
from typing import List, Union, Dict, Optional, Tuple
import os
import tempfile
from glob import glob
from pathlib import Path
import json

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq

import redvox.common.date_time_utils as dtu
from redvox.common import offset_model as om
from redvox.common.errors import RedVoxExceptions
from redvox.common.gap_and_pad_utils_wpa import calc_evenly_sampled_timestamps
from redvox.api1000.wrapped_redvox_packet.station_information import (
    NetworkType,
    PowerState,
    CellServiceState,
)
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
from redvox.api1000.wrapped_redvox_packet.sensors.image import ImageCodec
from redvox.api1000.wrapped_redvox_packet.sensors.audio import AudioCodec

# columns that cannot be interpolated
NON_INTERPOLATED_COLUMNS = ["compressed_audio", "image"]
# columns that are not numeric but can be interpolated
NON_NUMERIC_COLUMNS = ["location_provider", "image_codec", "audio_codec",
                       "network_type", "power_state", "cell_service"]


class SensorType(enum.Enum):
    """
    Enumeration of possible types of sensors to read data from
    """

    UNKNOWN_SENSOR = 0  # unknown sensor
    ACCELEROMETER = 1  # meters/second^2
    AMBIENT_TEMPERATURE = 2  # degrees Celsius
    AUDIO = 3  # normalized counts
    COMPRESSED_AUDIO = 4  # bytes (codec specific)
    GRAVITY = 5  # meters/second^2
    GYROSCOPE = 6  # radians/second
    IMAGE = 7  # bytes (codec specific)
    LIGHT = 8  # lux
    LINEAR_ACCELERATION = 9  # meters/second^2
    LOCATION = 10  # See standard
    MAGNETOMETER = 11  # microtesla
    ORIENTATION = 12  # radians
    PRESSURE = 13  # kilopascal
    PROXIMITY = 14  # on, off, cm
    RELATIVE_HUMIDITY = 15  # percentage
    ROTATION_VECTOR = 16  # Unitless
    INFRARED = 17  # this is proximity
    STATION_HEALTH = 18
    # battery charge and current level, phone internal temperature, network source and strength,
    # available RAM of the system, cell service status, amount of hard disk space left, power charging state
    BEST_LOCATION = 19  # See standard

    @staticmethod
    def type_from_str(type_str: str) -> "SensorType":
        """
        converts a string to a sensor type

        :param type_str: string to convert
        :return: a sensor type, UNKNOWN_SENSOR is the default for invalid inputs
        """
        if (
                type_str.lower() == "audio"
                or type_str.lower() == "mic"
                or type_str.lower() == "microphone"
        ):
            return SensorType.AUDIO
        elif type_str.lower() == "accelerometer" or type_str.lower() == "accel":
            return SensorType.ACCELEROMETER
        elif type_str.lower() == "ambient_temperature":
            return SensorType.AMBIENT_TEMPERATURE
        elif type_str.lower() == "compressed_audio":
            return SensorType.COMPRESSED_AUDIO
        elif type_str.lower() == "gravity":
            return SensorType.GRAVITY
        elif type_str.lower() == "gyroscope" or type_str.lower() == "gyro":
            return SensorType.GYROSCOPE
        elif type_str.lower() == "image":
            return SensorType.IMAGE
        elif type_str.lower() == "light":
            return SensorType.LIGHT
        elif (
                type_str.lower() == "linear_acceleration"
                or type_str.lower() == "linear_accel"
        ):
            return SensorType.LINEAR_ACCELERATION
        elif type_str.lower() == "location" or type_str.lower() == "loc":
            return SensorType.LOCATION
        elif type_str.lower() == "best_location" or type_str.lower() == "best_loc":
            return SensorType.BEST_LOCATION
        elif type_str.lower() == "magnetometer" or type_str.lower() == "mag":
            return SensorType.MAGNETOMETER
        elif type_str.lower() == "orientation":
            return SensorType.ORIENTATION
        elif (
                type_str.lower() == "pressure"
                or type_str.lower() == "bar"
                or type_str.lower() == "barometer"
        ):
            return SensorType.PRESSURE
        elif type_str.lower() == "proximity" or type_str.lower() == "infrared":
            return SensorType.PROXIMITY
        elif type_str.lower() == "relative_humidity":
            return SensorType.RELATIVE_HUMIDITY
        elif type_str.lower() == "rotation_vector":
            return SensorType.ROTATION_VECTOR
        else:
            return SensorType.UNKNOWN_SENSOR


class SensorDataPa:
    """
    Generic SensorData class for API-independent analysis

    Properties:
        name: string, name of sensor.  REQUIRED

        type: SensorType, enumerated type of sensor, default UNKNOWN_SENSOR

        sample_rate_hz: float, sample rate in Hz of the sensor, default np.nan, usually 1/sample_interval_s

        sample_interval_s: float, mean duration in seconds between samples, default np.nan, usually 1/sample_rate

        sample_interval_std_s: float, standard deviation in seconds between samples, default np.nan

        is_sample_rate_fixed: bool, True if sample rate is constant, default False

        timestamps_altered: bool, True if timestamps in the sensor have been altered from their original values
        default False

        use_offset_model: bool, if True, use an offset model to correct timestamps, otherwise use the best known
        offset.  default False

        errors: RedVoxExceptions, class containing a list of all errors encountered by the sensor.

        gaps: List of Tuples of floats, timestamps of data points on the edge of gaps, default empty list

    Protected:
        _arrow_file: string, file name of data file

        _arrow_dir: string, directory where the data is saved to, default "" (current directory)

        _save_data: bool, if True, save data to _arrow_dir, otherwise use a temp dir.  default False
    """

    def __init__(
            self,
            sensor_name: str,
            sensor_data: Optional[pa.Table] = None,
            sensor_type: SensorType = SensorType.UNKNOWN_SENSOR,
            sample_rate_hz: float = np.nan,
            sample_interval_s: float = np.nan,
            sample_interval_std_s: float = np.nan,
            is_sample_rate_fixed: bool = False,
            are_timestamps_altered: bool = False,
            calculate_stats: bool = False,
            use_offset_model_for_correction: bool = False,
            save_data: bool = False,
            arrow_dir: str = ".",
            gaps: Optional[List[Tuple[float, float]]] = None
    ):
        """
        initialize the sensor data with params

        :param sensor_name: name of the sensor.  REQUIRED
        :param sensor_type: enumerated type of the sensor, default SensorType.UNKNOWN_SENSOR
        :param sensor_data: Optional pyarrow table with the timestamps and sensor data;
                            first column is always the timestamps,
                            the other columns are the data channels in the sensor
                            if not given, will search arrow_dir for data
        :param sample_rate_hz: sample rate in hz of the data
        :param sample_interval_s: sample interval in seconds of the data
        :param sample_interval_std_s: std dev of sample interval in seconds of the data
        :param is_sample_rate_fixed: if True, sample rate is constant for all data, default False
        :param are_timestamps_altered: if True, timestamps in the sensor have been altered from their
                                        original values, default False
        :param calculate_stats: if True, calculate sample_rate, sample_interval_s, and sample_interval_std_s
                                default False
        :param use_offset_model_for_correction: if True, use an offset model to correct timestamps, otherwise
                                                use the best known offset.  default False
        :param save_data: if True, save the data of the sensor to disk, otherwise use a temporary dir.  default False
        :param arrow_dir: directory to save pyarrow table, default "." (current dir).  internally uses a temporary
                            dir if not saving data
        :param gaps: Optional list of timestamp pairs of data points on the edge of gaps in the data.  anything between
                        the pairs of points exists to maintain sample rate and are not considered valid points.
                        Default None
        """
        self.errors: RedVoxExceptions = RedVoxExceptions("Sensor")
        self.name: str = sensor_name
        self.type: SensorType = sensor_type
        self.sample_rate_hz: float = sample_rate_hz
        self.sample_interval_s: float = sample_interval_s
        self.sample_interval_std_s: float = sample_interval_std_s
        self.is_sample_rate_fixed: bool = is_sample_rate_fixed
        self.timestamps_altered: bool = are_timestamps_altered
        self.use_offset_model: bool = use_offset_model_for_correction
        self._arrow_file: str = ""
        self._save_data: bool = save_data
        if gaps:
            self.gaps = gaps
        else:
            self.gaps = []
        self._temp_dir = tempfile.TemporaryDirectory()
        self._arrow_dir: str = arrow_dir if self._save_data or arrow_dir else self._temp_dir.name
        self._data: Optional[pa.Table] = None
        if sensor_data:
            if "timestamps" not in sensor_data.schema.names:
                self.errors.append('must have a column titled "timestamps"')
            elif sensor_data['timestamps'].length() > 0:
                self._arrow_file: str = f"{sensor_type.name}_{int(sensor_data['timestamps'][0].as_py())}.parquet"
                if calculate_stats:
                    self.organize_and_update_stats(sensor_data)
                elif sensor_data["timestamps"].length() > 1:
                    self.sort_by_data_timestamps(sensor_data)
                else:
                    self.write_pyarrow_table(sensor_data)

    def __del__(self):
        self._temp_dir.cleanup()

    @staticmethod
    def from_dir(
            sensor_name: str,
            data_path: str,
            sensor_type: SensorType = SensorType.UNKNOWN_SENSOR,
            sample_rate_hz: float = np.nan,
            sample_interval_s: float = np.nan,
            sample_interval_std_s: float = np.nan,
            is_sample_rate_fixed: bool = False,
            are_timestamps_altered: bool = False,
            calculate_stats: bool = False,
            use_offset_model_for_correction: bool = False,
            save_data: bool = False) -> "SensorDataPa":
        """
        init but with a path to directory containing parquet file(s) instead of a table of data

        :param sensor_name: name of the sensor
        :param data_path: path to the directory containing the parquet files
        :param sensor_type: enumerated type of the sensor, default SensorType.UNKNOWN_SENSOR
        :param sample_rate_hz: sample rate in hz of the data, default np.nan
        :param sample_interval_s: sample interval in seconds of the data, default np.nan
        :param sample_interval_std_s: std dev of sample interval in seconds of the data, default np.nan
        :param is_sample_rate_fixed: if True, sample rate is constant for all data, default False
        :param are_timestamps_altered: if True, timestamps in the sensor have been altered from their
                                        original values, default False
        :param calculate_stats: if True, calculate sample_rate, sample_interval_s, and sample_interval_std_s
                                default False
        :param use_offset_model_for_correction: if True, use an offset model to correct timestamps, otherwise
                                                use the best known offset.  default False
        :param save_data: if True, save the data of the sensor to disk, otherwise use a temporary dir.  default False
        :return: SensorData object
        """
        result = SensorDataPa(sensor_name, None,
                              sensor_type, sample_rate_hz, sample_interval_s, sample_interval_std_s,
                              is_sample_rate_fixed, are_timestamps_altered, calculate_stats,
                              use_offset_model_for_correction, save_data, data_path)
        result.set_arrow_file()
        return result

    @staticmethod
    def from_dict(
            sensor_name: str,
            sensor_data: Dict,
            sensor_type: SensorType = SensorType.UNKNOWN_SENSOR,
            sample_rate_hz: float = np.nan,
            sample_interval_s: float = np.nan,
            sample_interval_std_s: float = np.nan,
            is_sample_rate_fixed: bool = False,
            are_timestamps_altered: bool = False,
            calculate_stats: bool = False,
            use_offset_model_for_correction: bool = False,
            save_data: bool = False,
            arrow_dir: str = "",
    ) -> "SensorDataPa":
        """
        init but with a dictionary

        :param sensor_name: name of the sensor
        :param sensor_type: enumerated type of the sensor, default SensorType.UNKNOWN_SENSOR
        :param sensor_data: dict with the timestamps and sensor data; first column is always the timestamps,
                            the other columns are the data channels in the sensor
        :param sample_rate_hz: sample rate in hz of the data, default np.nan
        :param sample_interval_s: sample interval in seconds of the data, default np.nan
        :param sample_interval_std_s: std dev of sample interval in seconds of the data, default np.nan
        :param is_sample_rate_fixed: if True, sample rate is constant for all data, default False
        :param are_timestamps_altered: if True, timestamps in the sensor have been altered from their
                                        original values, default False
        :param calculate_stats: if True, calculate sample_rate, sample_interval_s, and sample_interval_std_s
                                default False
        :param use_offset_model_for_correction: if True, use an offset model to correct timestamps, otherwise
                                                use the best known offset.  default False
        :param save_data: if True, save the data of the sensor to disk, otherwise use a temporary dir.  default False
        :param arrow_dir: directory to save pyarrow table, default "" (current dir).  default temporary dir if not
                            saving data
        :return: SensorData object
        """
        return SensorDataPa(sensor_name, pa.Table.from_pydict(sensor_data), sensor_type, sample_rate_hz,
                            sample_interval_s, sample_interval_std_s, is_sample_rate_fixed, are_timestamps_altered,
                            calculate_stats, use_offset_model_for_correction, save_data, arrow_dir)

    def is_save_to_disk(self) -> bool:
        """
        :return: True if sensor will be saved to disk
        """
        return self._save_data

    def set_arrow_file(self, new_file: Optional[str] = None):
        """
        set the pyarrow file name or use the default: {sensor_type.name}_{int(first_timestamp)}.parquet
        :param new_file: optional file name to change to; default None (use default name)
        """
        if new_file:
            self._arrow_file = new_file
        else:
            self._arrow_file = f"{self.type.name}_{int(self.first_data_timestamp())}.parquet"

    def set_arrow_dir(self, new_dir: str):
        """
        set the pyarrow directory
        :param new_dir: the directory to change to
        """
        self._arrow_dir = new_dir

    def pyarrow_ds(self) -> ds.Dataset:
        """
        :return: the dataset stored in self._arrow_dir
        """
        return ds.dataset(self._arrow_dir, format="parquet", exclude_invalid_files=True)

    def pyarrow_table(self) -> pa.Table:
        """
        :return: the table defined by the dataset stored in self._arrow_dir
        """
        if self._data:
            return self._data
        return self.pyarrow_ds().to_table()

    def data_df(self) -> pd.DataFrame:
        """
        :return: the pandas dataframe defined by the dataset stored in self._arrow_dir
        """
        return self.pyarrow_table().to_pandas()

    def write_pyarrow_table(self, table: pa.Table):
        """
        saves the pyarrow table to disk or to memory if self._save_data is False
        if writing to disk, uses a default filename: {sensor_type}_{first_timestamp}.parquet
        if writing to disk, removes the _data from memory.

        :param table: the table to write
        """
        if self._save_data:
            # clear the output dir where the table will be written to
            filelist = glob(os.path.join(self._arrow_dir, "*.parquet"))
            for f in filelist:
                os.remove(f)
            pq.write_table(table, os.path.join(self._arrow_dir, self._arrow_file))
            self._data = None
        else:
            self._data = table

    def sort_by_data_timestamps(self, ptable: pa.Table, ascending: bool = True):
        """
        sorts the data based on timestamps

        :param ptable: pyarrow table to sort
        :param ascending: if True, timestamps are sorted in ascending order, else sort by descending order
        """
        if ascending:
            order = "ascending"
        else:
            order = "descending"
        data = pc.take(ptable, pc.sort_indices(ptable, sort_keys=[("timestamps", order)]))
        self._arrow_file: str = f"{self.type.name}_{int(data['timestamps'][0].as_py())}.parquet"
        # self.write_pyarrow_table(data)
        self._data = data

    def organize_and_update_stats(self, ptable: pa.Table) -> "SensorDataPa":
        """
        sorts the data by timestamps, then if the sample rate is not fixed, recalculates the sample rate, interval,
            and interval std dev.  If there is only one value, sets the sample rate, interval, and interval std dev
            to np.nan.  Updates the SensorData object with the new values

        :param ptable: pyarrow table to update
        :return: updated version of self
        """
        self.sort_by_data_timestamps(ptable)
        if not self.is_sample_rate_fixed:
            if self.num_samples() > 1:
                timestamp_diffs = np.diff(self.data_timestamps())
                self.sample_interval_s = dtu.microseconds_to_seconds(
                    float(np.mean(timestamp_diffs))
                )
                self.sample_interval_std_s = dtu.microseconds_to_seconds(
                    float(np.std(timestamp_diffs))
                )
                self.sample_rate_hz = (
                    np.nan
                    if self.is_sample_interval_invalid()
                    else 1 / self.sample_interval_s
                )
            else:
                self.sample_interval_s = np.nan
                self.sample_interval_std_s = np.nan
                self.sample_rate_hz = np.nan
        return self

    def append_sensor(self, new_sensor: "SensorDataPa", recalculate_stats: bool = False) -> "SensorDataPa":
        """
        append the new data to the sensor, update the sensor's stats on demand if it doesn't have a fixed
            sample rate, then return the updated SensorData object

        :param new_sensor: sensor containing data to add to the calling sensor
        :param recalculate_stats: if True and the sensor does not have a fixed sample rate, sort the timestamps,
                                    recalculate the sample rate, interval, and interval std dev, default False
        :return: the updated SensorData object
        """
        _arrow: pa.Table = pa.concat_tables([self.pyarrow_table(), new_sensor.pyarrow_table()])
        if recalculate_stats and not self.is_sample_rate_fixed:
            self.organize_and_update_stats(_arrow)
        else:
            self.write_pyarrow_table(_arrow)
        return self

    def append_data(
            self, new_data: List[np.array], recalculate_stats: bool = False
    ) -> "SensorDataPa":
        """
        append the new data to the dataframe, update the sensor's stats on demand if it doesn't have a fixed
            sample rate, then return the updated SensorData object

        :param new_data: list of arrays containing data to add to the sensor's dataframe
        :param recalculate_stats: if True and the sensor does not have a fixed sample rate, sort the timestamps,
                                    recalculate the sample rate, interval, and interval std dev, default False
        :return: the updated SensorData object
        """
        _arrow = pa.concat_tables([self.pyarrow_table(),
                                   pa.Table.from_arrays(arrays=[pa.array(s) for s in new_data],
                                                        names=self.data_channels())])
        if recalculate_stats and not self.is_sample_rate_fixed:
            self.organize_and_update_stats(_arrow)
        else:
            self.write_pyarrow_table(_arrow)
        return self

    def is_sample_interval_invalid(self) -> bool:
        """
        :return: True if sample interval is np.nan or equal to 0.0
        """
        return np.isnan(self.sample_interval_s) or self.sample_interval_s == 0.0

    def sensor_type_as_str(self) -> str:
        """
        gets the sensor type as a string

        :return: sensor type of the sensor as a string
        """
        return self.type.name

    def data_timestamps(self) -> np.array:
        """
        :return: the timestamps as a numpy array
        """
        return self.pyarrow_table()["timestamps"].to_numpy()

    def unaltered_data_timestamps(self) -> np.array:
        """
        :return: the unaltered timestamps as a numpy array
        """
        return self.pyarrow_table()["unaltered_timestamps"].to_numpy()

    def first_data_timestamp(self) -> float:
        """
        :return: timestamp of the first data point
        """
        return self.data_timestamps()[0]

    def last_data_timestamp(self) -> float:
        """
        :return: timestamp of the last data point
        """
        return self.data_timestamps()[-1]

    def num_samples(self) -> int:
        """
        :return: the number of rows (samples) in the dataframe
        """
        if self._arrow_file:
            return len(self.data_timestamps())
        return 0

    def samples(self) -> np.ndarray:
        """
        gets the non-timestamp samples of dataframe

        :return: the data values of the dataframe as a numpy ndarray
        """
        return self.data_df().iloc[:, 2:].T.to_numpy()

    def data_channels(self) -> List[str]:
        """
        :return: a list of the names of the columns (data channels) of the dataframe
        """
        return self.pyarrow_table().schema.names

    def get_data_channel(self, channel_name: str) -> Union[np.array, List[str]]:
        """
        gets the data channel specified, raises an error and lists valid fields if channel_name is not in the dataframe

        :param channel_name: the name of the channel to get data for
        :return: the data values of the channel as a numpy array or list of strings for enumerated channels
        """
        _arrow = self.pyarrow_table()
        if channel_name not in _arrow.schema.names:
            self.errors.append(f"WARNING: {channel_name} does not exist; try one of {_arrow.schema.names}")
            return []
        if channel_name == "location_provider":
            return [LocationProvider(c).name for c in _arrow[channel_name]]
        elif channel_name == "image_codec":
            return [ImageCodec(c).name for c in _arrow[channel_name]]
        elif channel_name == "audio_codec":
            return [AudioCodec(c).name for c in _arrow[channel_name]]
        elif channel_name == "network_type":
            return [NetworkType(c).name for c in _arrow[channel_name]]
        elif channel_name == "power_state":
            return [PowerState(c).name for c in _arrow[channel_name]]
        elif channel_name == "cell_service":
            return [CellServiceState(c).name for c in _arrow[channel_name]]
        return _arrow[channel_name].to_numpy()

    def get_valid_data_channel_values(self, channel_name: str) -> np.array:
        """
        gets all non-nan values from the channel specified

        :param channel_name: the name of the channel to get data for
        :return: non-nan values of the channel as a numpy array
        """
        channel_data = self.get_data_channel(channel_name)
        return channel_data[~np.isnan(channel_data)]

    def update_data_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = True):
        """
        updates the timestamps of the data points

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the offset model's correction function to correct time,
                                    otherwise use best offset (model's intercept value).  default True
        """
        slope = dtu.seconds_to_microseconds(self.sample_interval_s) * (1 + offset_model.slope) \
            if use_model_function else dtu.seconds_to_microseconds(self.sample_interval_s)
        if self.type == SensorType.AUDIO:
            # use the model to update the first timestamp or add the best offset (model's intercept value)
            timestamps = pa.array(
                calc_evenly_sampled_timestamps(
                    offset_model.update_time(self.first_data_timestamp(), use_model_function),
                    self.num_samples(),
                    slope))
        else:
            timestamps = pa.array(offset_model.update_timestamps(self.data_timestamps(),
                                                                 use_model_function))
        self._data = self.pyarrow_table().set_column(0, "timestamps", timestamps)
        self.set_arrow_file()
        time_diffs = np.floor(np.diff(self.data_timestamps()))
        if len(time_diffs) > 1:
            self.sample_interval_s = dtu.microseconds_to_seconds(slope)
            if self.sample_interval_s > 0:
                self.sample_rate_hz = 1 / self.sample_interval_s
                self.sample_interval_std_s = dtu.microseconds_to_seconds(np.std(time_diffs))
        self.timestamps_altered = True

    def interpolate(self, interpolate_timestamp: float, first_point: int, second_point: int = 0,
                    copy: bool = True) -> pa.Table:
        """
        interpolates two points at the intercept value.  the two points must be consecutive in the dataframe.
        data channels that can't be interpolated are set to np.nan.

        :param interpolate_timestamp: timestamp to interpolate other values
        :param first_point: index of first point
        :param second_point: delta to second point, default 0 (same as first point)
        :param copy: if True, copies the values of the first point, default True
        :return: pyarrow Table of interpolated points
        """
        start_point = self.pyarrow_table().slice(first_point, 1).to_pydict()
        if not copy and second_point:
            i_p = {}
            end_point = self.pyarrow_table().slice(first_point + second_point, 1).to_pydict()
            first_closer = \
                np.abs(start_point[0] - interpolate_timestamp) \
                <= np.abs(end_point[0] - interpolate_timestamp)
            for col in self.pyarrow_table().schema.names:
                # process each column independently into new table object
                if col not in NON_INTERPOLATED_COLUMNS + NON_NUMERIC_COLUMNS:
                    numeric_diff = end_point[col] - start_point[col]
                    numeric_diff = \
                        (numeric_diff / numeric_diff["timestamps"]) * \
                        (interpolate_timestamp - start_point[col]) + start_point[col]
                    i_p[col] = numeric_diff
                elif col in NON_NUMERIC_COLUMNS:
                    if first_closer:
                        i_p[col] = start_point[col]
                    else:
                        i_p[col] = end_point[col]
        else:
            i_p = start_point
        i_p["timestamps"] = [interpolate_timestamp]
        return pa.Table.from_pydict(i_p)

    def get_file_path(self) -> str:
        """
        :return: the supposed file path to the data file
        """
        return os.path.join(self._arrow_dir, self._arrow_file)

    def as_dict(self) -> dict:
        """
        :return: sensor as dict
        """
        return {
            "name": self.name,
            "type": self.type.name,
            "num_samples": self.num_samples(),
            "sample_rate_hz": self.sample_rate_hz,
            "sample_interval_s": self.sample_interval_s,
            "sample_interval_std_s": self.sample_interval_std_s,
            "is_sample_rate_fixed": self.is_sample_rate_fixed,
            "timestamps_altered": self.timestamps_altered,
            "use_offset_model": self.use_offset_model,
            "gaps": self.gaps,
            "parquet_dir": self._arrow_dir,
            "parquet_file": self._arrow_file,
            "errors": self.errors.as_dict()
        }

    def to_json(self) -> str:
        """
        :return: sensor as json string
        """
        return json.dumps(self.as_dict())

    def to_json_file(self, file_name: Optional[str] = None) -> Path:
        """
        saves the sensor as json and data in the same directory.

        :param file_name: the optional base file name.  Do not include a file extension.
                            If None, a default file name is created using this format:
                            [sensor_type]_[first_timestamp].json
        :return: path to json file
        """
        _file_name: str = (
            file_name
            if file_name is not None
            else self._arrow_file.split(".")[0]
        )
        if self._save_data:
            pq.write_table(self._data, os.path.join(self._arrow_dir, self._arrow_file))
        file_path: Path = Path(self._arrow_dir).joinpath(f"{_file_name}.json")
        with open(file_path, "w") as f_p:
            f_p.write(self.to_json())
            return file_path.resolve(False)

    @staticmethod
    def from_json(file_path: str) -> "SensorDataPa":
        """
        convert contents of json file to SensorData
        :param file_path: full path of file to load data from.
        :return: SensorData object
        """
        with open(file_path, "r") as f_p:
            json_data = json.loads(f_p.read())
        if "name" in json_data.keys():
            result = SensorDataPa.from_dir(json_data["name"], json_data["parquet_dir"], SensorType[json_data["type"]],
                                           json_data["sample_rate_hz"], json_data["sample_interval_s"],
                                           json_data["sample_interval_std_s"], json_data["is_sample_rate_fixed"],
                                           json_data["timestamps_altered"], False, json_data["use_offset_model"])
            result.errors = RedVoxExceptions.from_dict(json_data["errors"])
            result.gaps = json_data["gaps"]

        else:
            result = SensorDataPa("Empty")
            result.errors.append("Loading from json file failed; Sensor missing name.")
        return result



"""
Read Redvox data from a single directory
Data files can be either API 900 or API 1000 data formats
"""
from typing import List, Optional
from datetime import timedelta
import multiprocessing
import multiprocessing.pool
from itertools import repeat
from multiprocessing import cpu_count, Manager, Process, Queue
import queue

import pyarrow as pa

import redvox.settings as settings
import redvox.api1000.proto.redvox_api_m_pb2 as api_m
from redvox.common import offset_model
from redvox.common import api_conversions as ac
from redvox.common import io
from redvox.common import file_statistics as fs
from redvox.common.parallel_utils import maybe_parallel_map
from redvox.common.station import Station
from redvox.common.station_wpa import StationPa
from redvox.common.errors import RedVoxExceptions


id_py_stct = pa.struct([("id", pa.string()), ("uuid", pa.string()), ("start_time", pa.float64()),
                        ])
meta_py_stct = pa.struct([("api", pa.float64()), ("sub_api", pa.float64()), ("make", pa.string()),
                          ("model", pa.string()), ("os", pa.int64()), ("os_version", pa.string()),
                          ("app", pa.string()), ("app_version", pa.string()), ("is_private", pa.bool_()),
                          ("packet_duration_s", pa.float64()), ("station_description", pa.string()),
                          ])


class ApiReader:
    """
    Reads data from api 900 or api 1000 format, converting all data read into RedvoxPacketM for
        ease of comparison and use.
    Properties:
        filter: io.ReadFilter with the station ids, start and end time, start and end time padding, and
                types of files to read
        base_dir: str of the directory containing all the files to read
        structured_dir: bool, if True, the base_dir contains a specific directory structure used by the
                        respective api formats.  If False, base_dir only has the data files.  Default False.
        files_index: io.Index of the files that match the filter that are in base_dir
        index_summary: io.IndexSummary of the filtered data
        debug: bool, if True, output additional information during function execution.  Default False.
    """

    def __init__(
        self,
        base_dir: str,
        structured_dir: bool = False,
        read_filter: io.ReadFilter = None,
        debug: bool = False,
        pool: Optional[multiprocessing.pool.Pool] = None,
    ):
        """
        Initialize the ApiReader object

        :param base_dir: directory containing the files to read
        :param structured_dir: if True, base_dir contains a specific directory structure used by the respective
                                api formats.  If False, base_dir only has the data files.  Default False.
        :param read_filter: ReadFilter for the data files, if None, get everything.  Default None
        :param debug: if True, output additional statements during function execution.  Default False.
        """
        _pool: multiprocessing.pool.Pool = (
            multiprocessing.Pool() if pool is None else pool
        )

        if read_filter:
            self.filter = read_filter
            if self.filter.station_ids:
                self.filter.station_ids = set(self.filter.station_ids)
        else:
            self.filter = io.ReadFilter()
        self.base_dir = base_dir
        self.structured_dir = structured_dir
        self.debug = debug
        self.errors = RedVoxExceptions("APIReader")
        self.files_index = self._get_all_files(_pool)
        self.index_summary = io.IndexSummary.from_index(self._flatten_files_index())

        if debug:
            self.errors.print()

        if pool is None:
            _pool.close()

    def _flatten_files_index(self):
        """
        :return: flattened version of files_index
        """
        result = io.Index()
        for i in self.files_index:
            result.append(i.entries)
        return result

    def _get_all_files(
        self, pool: Optional[multiprocessing.pool.Pool] = None
    ) -> List[io.Index]:
        """
        get all files in the base dir of the ApiReader

        :return: index with all the files that match the filter
        """
        _pool: multiprocessing.pool.Pool = (
            multiprocessing.Pool() if pool is None else pool
        )
        index: List[io.Index] = []
        # this guarantees that all ids we search for are valid
        all_index = self._apply_filter(pool=_pool)
        for station_id in all_index.summarize().station_ids():
            id_index = all_index.get_index_for_station_id(station_id)
            checked_index = self._check_station_stats(id_index, pool=_pool)
            index.extend(checked_index)

        if pool is None:
            _pool.close()

        return index

    def _apply_filter(
        self,
        reader_filter: Optional[io.ReadFilter] = None,
        pool: Optional[multiprocessing.pool.Pool] = None,
    ) -> io.Index:
        """
        apply the filter of the reader, or another filter if specified

        :param reader_filter: optional filter; if None, use the reader's filter, default None
        :return: index of the filtered files
        """
        _pool: multiprocessing.pool.Pool = (
            multiprocessing.Pool() if pool is None else pool
        )
        if not reader_filter:
            reader_filter = self.filter
        if self.structured_dir:
            index = io.index_structured(self.base_dir, reader_filter, pool=_pool)
        else:
            index = io.index_unstructured(self.base_dir, reader_filter, pool=_pool)
        if pool is None:
            _pool.close()
        return index

    def _check_station_stats(
            self,
            station_index: io.Index,
            pool: Optional[multiprocessing.pool.Pool] = None,
    ) -> List[io.Index]:
        """
        check the index's results; if it has enough information, return it, otherwise search for more data.
        The index should only request one station id
        If the station was restarted during the request period, a new group of indexes will be created
        to represent the change in station metadata.

        :param station_index: index representing the requested information
        :return: List of Indexes that includes as much information as possible that fits the request
        """
        _pool: multiprocessing.pool.Pool = multiprocessing.Pool() if pool is None else pool
        # if we found nothing, return the index
        if len(station_index.entries) < 1:
            return [station_index]

        stats = fs.extract_stats(station_index, pool=_pool)
        # Close pool if created here
        if pool is None:
            _pool.close()

        timing_offsets: Optional[offset_model.TimingOffsets] = offset_model.compute_offsets(stats)

        # punt if duration or other important values are invalid or if the latency array was empty
        if timing_offsets is None:
            return [station_index]

        diff_s = diff_e = timedelta(seconds=0)

        # if our filtered files do not encompass the request even when the packet times are updated
        # try getting 1.5 times the difference of the expected start/end and the start/end of the data
        insufficient_str = ""
        if self.filter.start_dt and timing_offsets.adjusted_start > self.filter.start_dt:
            insufficient_str += f" {self.filter.start_dt} (start)"
            # diff_s = self.filter.start_dt_buf + 1.5 * (timing_offsets.adjusted_start - self.filter.start_dt)
            new_end = self.filter.start_dt - self.filter.start_dt_buf
            new_start = new_end - 1.5 * (timing_offsets.adjusted_start - self.filter.start_dt)
            new_index = self._apply_filter(io.ReadFilter()
                                           .with_start_dt(new_start)
                                           .with_end_dt(new_end)
                                           .with_extensions(self.filter.extensions)
                                           .with_api_versions(self.filter.api_versions)
                                           .with_station_ids(set(station_index.summarize().station_ids()))
                                           .with_start_dt_buf(diff_s)
                                           .with_end_dt_buf(diff_e))
            if len(new_index.entries) > 0:
                station_index.append(new_index.entries)
                stats.extend(fs.extract_stats(new_index))
        if self.filter.end_dt and timing_offsets.adjusted_end < self.filter.end_dt:
            insufficient_str += f" {self.filter.end_dt} (end)"
            # diff_e = self.filter.end_dt_buf + 1.5 * (self.filter.end_dt - timing_offsets.adjusted_end)
            new_start = self.filter.end_dt + self.filter.end_dt_buf
            new_end = new_start + 1.5 * (self.filter.end_dt - timing_offsets.adjusted_end)
            new_index = self._apply_filter(io.ReadFilter()
                                           .with_start_dt(new_start)
                                           .with_end_dt(new_end)
                                           .with_extensions(self.filter.extensions)
                                           .with_api_versions(self.filter.api_versions)
                                           .with_station_ids(set(station_index.summarize().station_ids()))
                                           .with_start_dt_buf(diff_s)
                                           .with_end_dt_buf(diff_e))
            if len(new_index.entries) > 0:
                station_index.append(new_index.entries)
                stats.extend(fs.extract_stats(new_index))
        if len(insufficient_str) > 0:
            self.errors.append(f"Data for {station_index.summarize().station_ids()} exists, "
                               f"but not at:{insufficient_str}")

        results = {}
        keys = []

        for v, e in enumerate(stats):
            key = e.app_start_dt
            if key not in keys:
                keys.append(key)
                results[key] = io.Index()

            results[key].append(entries=[station_index.entries[v]])

        return list(results.values())

    @staticmethod
    def read_files_in_index(indexf: io.Index) -> List[api_m.RedvoxPacketM]:
        """
        read all the files in the index

        :return: list of RedvoxPacketM, converted from API 900 if necessary
        """
        result: List[api_m.RedvoxPacketM] = []

        # Iterate over the API 900 packets in a memory efficient way
        # and convert to API 1000
        # noinspection PyTypeChecker
        for packet_900 in indexf.stream_raw(
                io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_900})
        ):
            # noinspection Mypy
            result.append(
                ac.convert_api_900_to_1000_raw(packet_900)
            )

        # Grab the API 1000 packets
        # noinspection PyTypeChecker
        for packet in indexf.stream_raw(
                io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_1000})
        ):
            # noinspection Mypy
            result.append(packet)

        return result

    # noinspection PyTypeChecker
    def read_files_by_id(self, station_id: str) -> Optional[List[api_m.RedvoxPacketM]]:
        """
        :param station_id: the id to filter on
        :return: the list of packets with the requested id, or None if the id can't be found
        """

        result: List[api_m.RedvoxPacketM] = []

        # Iterate over the API 900 packets in a memory efficient way
        # and convert to API 1000
        for packet_900 in self._flatten_files_index().stream_raw(
            io.ReadFilter.empty()
            .with_api_versions({io.ApiVersion.API_900})
            .with_station_ids({station_id})
        ):
            # noinspection Mypy
            result.append(ac.convert_api_900_to_1000_raw(packet_900))

        # Grab the API 1000 packets
        for packet in self._flatten_files_index().stream_raw(
            io.ReadFilter.empty()
            .with_api_versions({io.ApiVersion.API_1000})
            .with_station_ids({station_id})
        ):
            # noinspection Mypy
            result.append(packet)

        if len(result) == 0:
            return None

        return result

    def _stations_by_index(self, findex: io.Index) -> Station:
        """
        :param findex: index with files to build a station with
        :return: Station built from files in findex
        """
        return Station(self.read_files_in_index(findex))

    def get_stations(self, pool: Optional[multiprocessing.pool.Pool] = None) -> List[Station]:
        """
        :param pool: optional multiprocessing pool
        :return: List of all stations in the ApiReader
        """
        return list(maybe_parallel_map(pool,
                                       self._stations_by_index,
                                       self.files_index,
                                       chunk_size=1
                                       )
                    )

    def get_station_by_id(self, get_id: str) -> Optional[List[Station]]:
        """
        :param get_id: the id to filter on
        :return: list of all stations with the requested id or None if id can't be found
        """
        result = [s for s in self.get_stations() if s.id == get_id]
        if len(result) < 1:
            return None
        return result

    def _stations_wpa_by_index_fs(self, findex: io.Index, correct_timestamps: bool = False,
                                  use_model_correction: bool = True,
                                  base_dir: str = "", save_files: bool = False) -> StationPa:
        """
        :param findex: index with files to build a station with
        :param correct_timestamps: if True, correct timestamps as soon as they're available.  Default False
        :param use_model_correction: if True, use OffsetModel functions for time correction, add OffsetModel
                                        best offset (intercept value) otherwise.  Default True
        :param base_dir: base directory to write data parquet files to.  Default "" (current directory)
        :param save_files: if True, save files to disk, otherwise delete when finished.  Default False
        :return: Station built from files in findex, without building the data from parquet
        """
        stpa = StationPa.create_from_packets(self.read_files_in_index(findex), correct_timestamps,
                                             use_model_correction, base_dir, save_files)
        return stpa

    def download_process(self, input_queue: Queue, result_queue: Queue, correct_timestamps: bool = False,
                         use_model_correction: bool = True, base_dir: str = "", save_files: bool = False
                         ) -> None:
        """
        A function that runs in a separate process for creating stations.
        :param input_queue: A shared queue containing the list of items to be downloaded.
        :param result_queue: A queue used to send results from the process to the caller.
        :param correct_timestamps: if True, correct timestamps as soon as they're available.  Default False
        :param use_model_correction: if True, use OffsetModel functions for time correction, add OffsetModel
                                        best offset (intercept value) otherwise.  Default True
        :param base_dir: base directory to write data parquet files to.  Default "" (current directory)
        :param save_files: if True, save files to disk, otherwise delete when finished.  Default False
        """
        try:
            # While there is still data in the queue, retrieve it.
            while True:
                files_index = input_queue.get_nowait()
                try:
                    st = StationPa.create_from_packets(self.read_files_in_index(files_index), correct_timestamps,
                                                       use_model_correction, base_dir, save_files)
                    result_queue.put(st, True, None)
                except FileExistsError:
                    print(f"File already exists, skipping...")
                    continue
        # Thrown when the queue is empty
        except queue.Empty:
            return

    def download_files(self, num_processes: int = cpu_count(), correct_timestamps: bool = False,
                       use_model_correction: bool = True, base_dir: str = "", save_files: bool = False):
        """
        Create stations in parallel.
        :param num_processes: Number of processes to create for downloading data.
        :param correct_timestamps: if True, correct timestamps as soon as possible, default False
        :param use_model_correction: if True, use OffsetModel functions for time correction, add OffsetModel
                                        best offset (intercept value) otherwise.  Default True
        :param base_dir: base directory to write data parquet files to.  Default "" (current directory)
        :param save_files: if True, save files to disk, otherwise delete when finished.  Default False
        """
        manager: Manager = Manager()
        station_queue: Queue = manager.Queue(len(self.files_index))
        result_queue: Queue = manager.Queue(len(self.files_index))
        processes: List[Process] = []

        # Add all URLs to shared queue
        for findex in self.files_index:
            station_queue.put(findex)

        # Create the process pool
        for _ in range(num_processes):
            process: Process = Process(
                target=self.download_process, args=(station_queue, result_queue, correct_timestamps,
                                                    use_model_correction, base_dir, save_files)
            )
            processes.append(process)
            process.start()

        # Wait for all processes in pool to finish
        for process in processes:
            process.join()

        result = []
        try:
            while True:
                x = result_queue.get_nowait()
                result.append(x)
        except queue.Empty:
            return result

    def get_stations_wpa_fs(self,
                            # pool: Optional[multiprocessing.pool.Pool] = None,
                            correct_timestamps: bool = False,
                            use_model_correction: bool = True,
                            base_dir: str = "", save_files: bool = False) -> List[StationPa]:
        """
        # :param pool: optional multiprocessing pool
        :param correct_timestamps: if True, correct timestamps as soon as possible, default False
        :param use_model_correction: if True, use OffsetModel functions for time correction, add OffsetModel
                                        best offset (intercept value) otherwise.  Default True
        :param base_dir: base directory to write data parquet files to.  Default "" (current directory)
        :param save_files: if True, save files to disk, otherwise delete when finished.  Default False
        :return: List of all stations in the ApiReader, without building the data from parquet
        """
        # return list(maybe_parallel_smap(pool,
        #                                 self._stations_wpa_by_index_fs,
        #                                 [self.files_index, repeat(base_dir), repeat(save_files)],
        #                                 chunk_size=1
        #                                 )
        #             )
        if settings.is_parallelism_enabled():
            return self.download_files(correct_timestamps=correct_timestamps,
                                       use_model_correction=use_model_correction,
                                       base_dir=base_dir, save_files=save_files)
        return list(map(self._stations_wpa_by_index_fs, self.files_index, repeat(correct_timestamps),
                        repeat(use_model_correction), repeat(base_dir), repeat(save_files)))

    def _stations_wpa_by_index(self, findex: io.Index) -> StationPa:
        """
        :param findex: index with files to build a station with
        :return: Station built from files in findex
        """
        stpa = StationPa.create_from_packets(self.read_files_in_index(findex))
        return stpa

    def get_stations_wpa(self, pool: Optional[multiprocessing.pool.Pool] = None) -> List[StationPa]:
        """
        :param pool: optional multiprocessing pool
        :return: List of all stations in the ApiReader
        """
        return list(maybe_parallel_map(pool,
                                       self._stations_wpa_by_index,
                                       self.files_index,
                                       chunk_size=1
                                       )
                    )

    def get_station_wpa_by_id(self, get_id: str) -> Optional[List[StationPa]]:
        """
        :param get_id: the id to filter on
        :return: list of all stations with the requested id or None if id can't be found
        """
        result = [s for s in self.get_stations_wpa() if s.get_id() == get_id]
        if len(result) < 1:
            return None
        return result
